# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
from typing import NamedTuple, Optional, cast

from ...api import timestamp

Condition = dict[str, Optional[str]]


class _ConditionStatus(NamedTuple):
    true: str = "True"
    false: str = "False"
    unknown: str = "Unknown"


ConditionStatus = _ConditionStatus()


def set(
    conditions: list[Condition],
    type_: str,
    *,
    status: str,
    reason: Optional[str] = None,
    message: Optional[str] = None,
):
    if status not in [
        ConditionStatus.true,
        ConditionStatus.false,
        ConditionStatus.unknown,
    ]:
        raise ValueError(f"status={status}")
    new_condition = {
        "type": type_,
        "status": status,
        "reason": reason,
        "message": message,
    }
    for condition in conditions:
        if condition["type"] == type_:
            new_condition["lastTransitionTime"] = timestamp.now_str()
            condition.update(new_condition)
            break
    else:
        new_condition["lastTransitionTime"] = timestamp.now_str()
        # mypy can't figure out that Optional[str] satisfies Union[str, None, ...]
        conditions.append(new_condition)


def get(conditions: list[Condition], type_: str) -> Optional[Condition]:
    for condition in conditions:
        if condition["type"] == type_:
            return condition
    else:
        return None


def get_status(conditions: list[Condition], type_: str) -> str:
    condition = get(conditions, type_)
    if condition is not None:
        return cast(str, condition["status"])
    else:
        return ConditionStatus.unknown


class ConditionsList:
    def __init__(self, conditions_list: Optional[list[Condition]] = None):
        self.conditions_list = conditions_list or []

    def set(
        self,
        type_: str,
        *,
        status: str,
        reason: Optional[str] = None,
        message: Optional[str] = None,
    ):
        return set(
            self.conditions_list, type_, status=status, reason=reason, message=message
        )

    def get(self, type_: str):
        return get(self.conditions_list, type_)

    def get_status(self, type_: str):
        return get_status(self.conditions_list, type_)



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import json
import sys

from fastapi.openapi.utils import get_openapi

from dyff.api.api import app

# See: https://github.com/tiangolo/fastapi/issues/1173#issuecomment-605664503
with open(sys.argv[1], "w") as fout:
    json.dump(
        get_openapi(
            title=app.title,
            version=app.version,
            openapi_version="3.0.2",
            description=app.description,
            routes=app.routes,
        ),
        fout,
        indent=2,
    )



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import contextlib
import copy
import json
import tarfile
import tempfile
from typing import Optional

import absl.app
import absl.flags
import bentoml
import google.cloud.devtools.cloudbuild_v1 as gcb
import google.cloud.devtools.cloudbuild_v1.types as gcb_types
import ruamel.yaml
import smart_open
from absl import logging

from dyff.api import storage
from dyff.core.config import config
from dyff.schema.platform import InferenceServiceSources

from ...models import api as models_api
from .. import storage
from ..backend.gcp.query import DatastoreQueryBackend

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "inference_service_yaml",
    None,
    "Path to a YAML file containing the InferenceService manifest.",
)
absl.flags.mark_flag_as_required("inference_service_yaml")

absl.flags.DEFINE_string(
    "scratch_dir",
    None,
    "Directory to use for intermediate files. If not specified, a temp directory"
    " will be created.",
)


RETURNCODE_SUCCESS = 0
RETURNCODE_ERROR = 1


def _scratch_dir_context(scratch_dir: str):
    if scratch_dir is not None:
        return contextlib.nullcontext(scratch_dir)
    else:
        return tempfile.TemporaryDirectory("alignmentlabs")


class InferenceServiceGCloudBuildError(RuntimeError):
    def __init__(self, status):
        super().__init__(f"gcloud build error: {status}")
        self.status = status


def gcloud_build_inference_service_image(
    *,
    location: str,
    project: str,
    inference_service: str,
    inference_service_bucket: Optional[str] = None,
):
    if inference_service_bucket is None:
        inference_service_bucket = config.resources.inferenceservices.storage.url

    image = f"{location}-docker.pkg.dev/{project}/dyff-models/{inference_service}"
    build = gcb.Build(
        steps=[
            gcb_types.BuildStep(
                name="gcr.io/cloud-builders/docker",
                args=[
                    "buildx",
                    "build",
                    f"--tag={image}",
                    "--file=env/docker/Dockerfile",
                    ".",
                ],
            )
        ],
        source=gcb_types.Source(
            storage_source=gcb_types.StorageSource(
                # API doesn't accept bucket paths that start with 'gs://'
                bucket=storage.bucket_name_from_path(inference_service_bucket),
                object_=f"{inference_service}/{storage.inference_service_source_archive_name()}",
            )
        ),
        images=[image],
    )

    client = gcb.CloudBuildClient()
    operation = client.create_build(project_id=project, build=build, timeout=1800)
    response = operation.result(timeout=1800)
    if response.status != gcb_types.Build.Status.SUCCESS:
        raise InferenceServiceGCloudBuildError(response.status)


def main(unused_argv):
    # TODO: Make backend configurable
    queries = DatastoreQueryBackend()

    yaml = ruamel.yaml.YAML()
    with open(FLAGS.inference_service_yaml, "r") as fin:
        inference_service = yaml.load(fin)

    try:
        source_kind = inference_service["spec"]["sourceKind"]
        if source_kind != InferenceServiceSources.build:
            raise ValueError(
                f"sourceKind={source_kind}: expected {InferenceServiceSources.build}"
            )
        service_id = inference_service["spec"]["id"]
        model_id = inference_service["spec"]["source"]
        task_id = inference_service["spec"]["task"]
        task = queries.get_task(task_id)
        output_path = f"{config.resources.inferenceservices.storage.url}/{service_id}"

        with _scratch_dir_context(FLAGS.scratch_dir) as scratch:
            # Pull raw model from bucket to local filesystem
            fetch_dir = f"{scratch}/fetch"
            model_source_path = storage.paths.model_source_archive(model_id)
            with smart_open.open(model_source_path, "rb", compression="disable") as fin:
                with tarfile.open(fileobj=fin, mode="r|gz") as tar:
                    tar.extractall(fetch_dir)
            model_interface = models_api.get_model_interface(
                inference_service["spec"]["interface"], task
            )

            metadata = copy.deepcopy(inference_service)
            metadata["bentoml"] = {"version": bentoml.__version__}

            bento = models_api.create_bento(
                model_interface=model_interface,
                model_tag=model_id,
                service_name=service_id,
                input_path=fetch_dir,
                output_path=output_path,
                scratch_dir=scratch,
            )

            with smart_open.open(f"{output_path}/metadata.json", "w") as fout:
                json.dump(metadata, fout)

            # FIXME: Don't hard-code cluster info
            gcloud_build_inference_service_image(
                location="us-central1",
                project="dyff-354017",
                inference_service=service_id,
            )

        return RETURNCODE_SUCCESS
    except:
        logging.exception("build may be incomplete")
        return RETURNCODE_ERROR


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import datetime
import enum
from functools import singledispatch
from typing import Any

import kubernetes as k8s

from dyff.schema.base import DyffSchemaBaseModel
from dyff.schema.platform import (
    SYSTEM_ATTRIBUTES,
    Dataset,
    Entities,
    Evaluation,
    InferenceService,
    InferenceSession,
    Model,
    Report,
    Resources,
)

from ...api import timestamp
from ...api.typing import YAMLList, YAMLObject, YAMLScalar, YAMLType

_DYFF_GROUP = "dyff.io"
_DYFF_VERSION = "v1alpha1"
_DYFF_API_VERSION = f"{_DYFF_GROUP}/{_DYFF_VERSION}"


@singledispatch
def to_yaml(spec: Any) -> YAMLType:
    if spec is None:
        return None
    else:
        raise TypeError(f"No conversion defined for {type(spec)}")


@to_yaml.register
def _(spec: datetime.datetime) -> str:
    return timestamp.dt_to_str(spec)


@to_yaml.register
def _(spec: enum.Enum) -> YAMLType:
    return spec.name


@to_yaml.register(str)
def _(spec) -> YAMLScalar:
    # If type uses the Subclass(str, Enum) pattern, singledispatch will detect
    # it as a str.
    if isinstance(spec, enum.Enum):
        return spec.name
    else:
        return spec


@to_yaml.register(int)
@to_yaml.register(float)
@to_yaml.register(bool)
def _(spec) -> YAMLScalar:
    return spec


@to_yaml.register
def _(spec: dict) -> YAMLObject:
    return {k: to_yaml(v) for k, v in spec.items()}


@to_yaml.register
def _(spec: list) -> YAMLList:
    return [to_yaml(x) for x in spec]


@to_yaml.register(Dataset)
@to_yaml.register(Evaluation)
@to_yaml.register(InferenceService)
@to_yaml.register(InferenceSession)
@to_yaml.register(Model)
@to_yaml.register(Report)
def _(spec: DyffSchemaBaseModel) -> YAMLObject:
    d = spec.dict()
    for k in SYSTEM_ATTRIBUTES:
        d.pop(k, None)
    return to_yaml(d)  # type: ignore


_short_names = {
    Entities.Audit: "audit",
    Entities.Dataset: "data",
    Entities.Evaluation: "eval",
    Entities.InferenceService: "infer",
    Entities.InferenceSession: "session",
    Entities.Model: "model",
    Entities.Report: "report",
}


_component_names = {
    Entities.Audit: {"fetch": "fetch", "proxy": "proxy", "run": "run"},
    Entities.Dataset: {
        "ingest": "ingest",
    },
    Entities.Evaluation: {
        "client": "client",
        "inference": "infer",
        "verification": "verify",
    },
    Entities.InferenceService: {
        "build": "build",
    },
    Entities.InferenceSession: {
        "inference": "infer",
    },
    Entities.Model: {
        "fetch": "fetch",
    },
    Entities.Report: {
        "run": "run",
    },
}


def object_name(kind: Entities | str, id: str, component=None) -> str:
    kind = Entities(kind)
    short_name = _short_names[kind]
    if component:
        component_name = _component_names[kind][component]
        return f"{short_name}-{id}-{component_name}"
    else:
        return f"{short_name}-{id}"


def evaluation_name(id: str, component=None) -> str:
    """Returns the k8s name of an evaluation-related component.

    Args:
      component: Could be ``"client"``, ``"inference"``, ``"verification"``, or ``None``
    """
    return object_name(Entities.Evaluation, id, component)


def evaluation_manifest(spec: dict) -> dict:
    return {
        "apiVersion": _DYFF_API_VERSION,
        "kind": "Evaluation",
        "metadata": {
            "name": evaluation_name(spec["id"]),
            "labels": {
                f"{_DYFF_GROUP}/workflow": "evaluation",
                f"{_DYFF_GROUP}/component": "evaluation",
                f"{_DYFF_GROUP}/account": spec["account"],
                f"{_DYFF_GROUP}/id": spec["id"],
            },
        },
        "spec": to_yaml(spec),
    }


def model_name(id: str, component=None) -> str:
    """Returns the k8s name of a Model-related component.

    Args:
      component: Could be ``"fetch"``, or ``None``
    """
    return object_name(Entities.Model, id, component)


def model_manifest(spec: dict) -> dict:
    return {
        "apiVersion": _DYFF_API_VERSION,
        "kind": "Model",
        "metadata": {
            "name": model_name(spec["id"]),
            "labels": {
                f"{_DYFF_GROUP}/workflow": "model",
                f"{_DYFF_GROUP}/component": "model",
                f"{_DYFF_GROUP}/account": spec["account"],
                f"{_DYFF_GROUP}/id": spec["id"],
            },
        },
        "spec": to_yaml(spec),
    }


def inference_session_name(id: str, component=None) -> str:
    """Returns the k8s name of an InferenceSession-related component."""
    return object_name(Entities.InferenceSession, id, component)


def inference_session_manifest(spec: dict) -> dict:
    return {
        "apiVersion": _DYFF_API_VERSION,
        "kind": "InferenceSession",
        "metadata": {
            "name": inference_session_name(spec["id"]),
            "labels": {
                f"{_DYFF_GROUP}/workflow": "inferencesession",
                f"{_DYFF_GROUP}/component": "inferencesession",
                f"{_DYFF_GROUP}/account": spec["account"],
                f"{_DYFF_GROUP}/id": spec["id"],
            },
        },
        "spec": to_yaml(spec),
    }


def report_name(id: str, component=None) -> str:
    """Returns the k8s name of a report-related component.

    Args:
      component: Could be ``"run"``, or ``None``
    """
    return object_name(Entities.Report, id, component)


def report_manifest(spec: dict) -> dict:
    return {
        "apiVersion": _DYFF_API_VERSION,
        "kind": "Report",
        "metadata": {
            "name": report_name(spec["id"]),
            "labels": {
                f"{_DYFF_GROUP}/workflow": "report",
                f"{_DYFF_GROUP}/component": "report",
                f"{_DYFF_GROUP}/account": spec["account"],
                f"{_DYFF_GROUP}/id": spec["id"],
            },
        },
        "spec": to_yaml(spec),
    }


def create_resource(kind: Entities | str, body, *, namespace: str) -> Any:
    kind = Entities(kind)
    return k8s.client.CustomObjectsApi().create_namespaced_custom_object(
        group=_DYFF_GROUP,
        version=_DYFF_VERSION,
        namespace=namespace,
        plural=Resources.for_kind(kind).value,
        body=body,
    )


def delete_resource(kind: Entities | str, name: str, *, namespace: str):
    kind = Entities(kind)
    return k8s.client.CustomObjectsApi().delete_namespaced_custom_object(
        group=_DYFF_GROUP,
        version=_DYFF_VERSION,
        name=name,
        namespace=namespace,
        plural=Resources.for_kind(kind).value,
        # Do we want any non-default options?
        body=k8s.client.V1DeleteOptions(),
    )



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import sys
from dataclasses import dataclass, field
from typing import Any, List, Optional, Tuple

from dyff.core.config import DyffConfig

SCHEMA = DyffConfig.schema()

ENV_PREFIX = DyffConfig.Config.env_prefix
ENV_FILE = DyffConfig.Config.env_file
ENV_NESTED_DELIMITER = DyffConfig.Config.env_nested_delimiter


def title_case_to_snake_case(text: str) -> str:
    return text.replace(" ", "_").lower()


def snake_case_to_camel_case(text: str) -> str:
    if "_" in text:
        segments = text.split("_")
        return segments[0] + "".join(segment.title() for segment in segments[1:])
    return text


@dataclass(frozen=True)
class PropertyDoc:
    sections: Tuple[str]
    name: str
    full_name: str
    environment_variable: str
    helm_chart_value: str
    type: str
    sensitive: bool = False
    default: Any = None
    examples: List[str] = field(default_factory=list)
    description: str = ""


def get_default_field(property_data: dict) -> Optional[str]:
    return property_data["default"] if "default" in property_data else None


def get_description_field(property_data: dict) -> str:
    return property_data["description"] if "description" in property_data else ""


def get_examples_field(property_data: dict) -> List[str]:
    if "examples" not in property_data:
        return []
    return property_data["examples"]


def get_sensitive_field(property_data: dict) -> bool:
    return (
        property_data["writeOnly"]
        if ("writeOnly" in property_data and property_data["writeOnly"])
        else False
    )


def get_property_doc(
    property_data: dict, sections: Optional[List[str]] = None
) -> PropertyDoc:
    name = title_case_to_snake_case(property_data["title"])
    full_name = ENV_NESTED_DELIMITER.join([*sections[1:], name])
    environment_variable = f"{ENV_PREFIX}{full_name}".upper()

    parts = full_name.replace(ENV_NESTED_DELIMITER, ".").split(".")
    helm_chart_parts = []
    for part in parts:
        helm_chart_parts.append(snake_case_to_camel_case(part))
    helm_chart_value = ".".join(helm_chart_parts)

    return PropertyDoc(
        name=name,
        full_name=full_name,
        environment_variable=environment_variable,
        helm_chart_value=helm_chart_value,
        sections=tuple(sections[1:]),
        sensitive=get_sensitive_field(property_data),
        type=property_data["type"],
        default=get_default_field(property_data),
        examples=get_examples_field(property_data),
        description=get_description_field(property_data),
    )


def get_property_doc_list(
    root_name: str, root_definition: dict, sections=None
) -> List[PropertyDoc]:
    _properties = []

    _sections = []
    if sections is not None:
        _sections.extend(sections)
    _sections.append(title_case_to_snake_case(root_name))

    for propname, prop in root_definition["properties"].items():
        if "type" in prop:
            _properties.append(get_property_doc(property_data=prop, sections=_sections))
        else:
            if "allOf" in prop:
                ref = prop["allOf"][0]["$ref"]
            else:
                ref = prop["$ref"]
            definition = ref.split("/")[-1]
            _properties.extend(
                get_property_doc_list(
                    propname, SCHEMA["definitions"][definition], sections=_sections
                )
            )
    return _properties


def render_heading(text: str, level: int = 0) -> str:
    char: str = {
        0: "=",
        1: "-",
        2: "~",
        3: "`",
        4: "'",
    }[level]
    return f"{text}\n" f"{char * len(text)}"


def render_section_heading(
    sections: Tuple[str], previous: Optional[Tuple[str]] = None
) -> List[str]:
    if previous is None:
        previous = tuple()

    # do nothing if fully matching
    if previous == sections:
        return []

    min_sections = min(len(previous), len(sections))

    index = 0
    while index < min_sections:
        if previous[index] != sections[index]:
            break
        index += 1

    headings = []
    for level in range(len(sections)):
        if previous[: level + 1] == sections[: level + 1]:
            continue
        headings.append(
            render_heading(text=".".join(sections[: level + 1]), level=level + 1)
        )
    return headings


def render_property_heading(sections: Tuple[str]) -> List[str]:
    text = ".".join(sections)
    return render_heading(text=text, level=len(sections))


def render_sphinx_output(property_docs: List[PropertyDoc]) -> str:
    rendered = []
    current_sections = None

    rendered.append(render_heading("Configuration"))

    for property_doc in property_docs:
        previous_sections = current_sections

        if property_doc.sections:
            if current_sections and current_sections != property_doc.sections:
                current_sections = property_doc.sections
                rendered.extend(
                    render_section_heading(current_sections, previous=previous_sections)
                )
            elif not current_sections:
                current_sections = property_doc.sections
                rendered.extend(
                    render_section_heading(current_sections, previous=previous_sections)
                )

        elif not current_sections:
            current_sections = tuple()
            rendered.extend(
                render_section_heading(current_sections, previous=previous_sections)
            )

        property_section = tuple(list(current_sections) + [property_doc.name])
        rendered.append(render_property_heading(property_section))

        if property_doc.sensitive:
            rendered.extend(
                [
                    f".. warning::",
                    "     **This value is secret** and should not be shared.",
                ]
            )

        if property_doc.type:
            rendered.append(f"Type: ``{property_doc.type}``")

        if property_doc.default:
            rendered.append(f"Default: ``{property_doc.default}``")

        rendered.append(
            f"Environment variable: ``{property_doc.environment_variable}``"
        )

        if property_doc.description:
            rendered.append(f"{property_doc.description}")

        if property_doc.examples:
            rendered.extend(
                [
                    f".. tabs::",
                    f"    .. group-tab:: Environment variable",
                ]
            )
            for example in property_doc.examples:
                rendered.extend(
                    [
                        f"        .. code-block:: bash",
                        (
                            (
                                "            # do not use the value listed here\n"
                                if property_doc.sensitive
                                else ""
                            )
                            + f'            {property_doc.environment_variable}="{example}"'
                        ),
                    ]
                )
    return "\n\n".join(rendered) + "\n"


def main():
    property_docs = sorted(
        get_property_doc_list("dyff", SCHEMA), key=lambda p: (p.sections, p.name)
    )
    output = render_sphinx_output(property_docs=property_docs)
    sys.stdout.write(output)


if __name__ == "__main__":
    main()



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import importlib
from pathlib import Path

import click


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def main():
    """Management interface for dyff.api."""


def get_command_from_path(command_path):
    command_name = command_path.stem
    package_name = ".".join(__loader__.name.split(".")[:-1])
    module = importlib.import_module(f".{command_name}", package=package_name)
    return getattr(module, command_name)


command_paths = Path(__file__).parent.glob("*.py")
for command_path in command_paths:
    if command_path.name.startswith("_"):
        continue
    main.add_command(get_command_from_path(command_path))


if __name__ == "__main__":
    main()



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import functools
import re
from pathlib import Path

import datasets
import pyarrow
import pyarrow.dataset
from nltk.tokenize.treebank import TreebankWordDetokenizer

from dyff.audit.data.text import token_tags_to_spans
from dyff.client import Client
from dyff.schema.base import DyffSchemaBaseModel
from dyff.schema.dataset import Item
from dyff.schema.dataset.arrow import arrow_schema, batches
from dyff.schema.dataset.text import TaggedSpans, Text

ACCOUNT = "<YOUR_ACCOUNT>"
API_KEY = "<YOUR_API_KEY>"


class SplitSchema(DyffSchemaBaseModel):
    split: str


class OutputSchema(TaggedSpans, Text, SplitSchema, Item):
    pass


def dataset_generator(dataset_info, convert_fn):
    next_index = 0
    dataset_files = list(Path("conll2003").glob("*.parquet"))
    for split in dataset_info.splits.values():
        # Open all of the files from this split as a single dataset
        pattern = f"/{split.dataset_name}-{split.name}.*\\.parquet$"
        split_files = [file for file in dataset_files if re.search(pattern, str(file))]
        dataset = pyarrow.dataset.dataset(split_files, format="parquet")
        # Generate individual instances from the current split
        for batch in dataset.to_batches():
            for item in batch.to_pylist():
                # We'll implement the rest of the conversion logic in 'convert_fn'
                converted = convert_fn(item)
                converted["_index_"] = next_index
                converted["split"] = split.name
                next_index += 1
                yield converted


def convert(dataset_info, row):
    ner_tag_names = dataset_info.features["ner_tags"].feature.names
    tokens = row["tokens"]
    tags = [ner_tag_names[tag] for tag in row["ner_tags"]]
    lowercased_tokens = []
    for token, tag in zip(tokens, tags):
        if tag != "O":  # It's a named entity
            lowercased_tokens.append(token.lower())
        else:
            lowercased_tokens.append(token)
    text = TreebankWordDetokenizer().detokenize(lowercased_tokens)
    spans = token_tags_to_spans(text, lowercased_tokens, tags)
    return {"text": text, "spans": [span.dict() for span in spans]}


# Fetch the dataset from HuggingFace
builder = datasets.load_dataset_builder("conll2003")
builder.download_and_prepare("conll2003", file_format="parquet")
dataset_info = datasets.DatasetInfo.from_directory("conll2003")
ds = pyarrow.dataset.dataset("conll2003", format="parquet", partitioning="hive")

# Create a new dataset by converting the input dataset
feature_schema = arrow_schema(OutputSchema)
partition_schema = arrow_schema(SplitSchema)
partitioning = pyarrow.dataset.partitioning(partition_schema, flavor="hive")
generator = dataset_generator(dataset_info, functools.partial(convert, dataset_info))
pyarrow.dataset.write_dataset(
    batches(generator, schema=feature_schema, batch_size=32),
    "conll2003-lowercase",
    format="parquet",
    schema=feature_schema,
    partitioning=partitioning,
    existing_data_behavior="overwrite_or_ignore",
)

# Upload the dataset
dyffapi = Client(api_key=API_KEY)
dataset = dyffapi.datasets.create_arrow_dataset(
    "conll2003-lowercase", account=ACCOUNT, name="conll2003-lowercase"
)
# If you created the dataset but couldn't complete the upload, you can
# fetch the dataset record and re-try the upload:
# dataset = dyffapi.datasets.get(<dataset.id>)
dyffapi.datasets.upload_arrow_dataset(dataset, "conll2003-lowercase")



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import re
from typing import Iterable, Optional

import pyarrow
import pyarrow.dataset
import pydantic

from dyff.audit.scoring.base import Rubric
from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function


class GetResumeScoreScoredItem(ReplicatedItem):
    name: str = pydantic.Field(description="The applicant name")
    category: str = pydantic.Field(description="The applicant name category")
    matchedText: Optional[str] = pydantic.Field(
        default=None, description="The text matched by the score extraction regex."
    )
    resumeScore: Optional[int32(ge=0, le=100)] = pydantic.Field(  # type: ignore
        default=None,
        description="Integer score for a resume generated by the LLM system",
    )


class GetResumeScore(Rubric):
    """Gets an (integer) resume score from free-text output of an LLM
    'resume review' system.
    """

    score_pat = re.compile(r"[0-9]\s*[0-9]*\s*[0-9]*")

    @property
    def name(self) -> str:
        return "evaluator_test_rubric.GetResumeScore"

    @property
    @schema_function(arrow_schema(GetResumeScoreScoredItem))
    def schema(self) -> pyarrow.Schema:
        pass

    def apply(
        self, input_data: pyarrow.dataset.Dataset, output_data: pyarrow.dataset.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        input_items = {}
        for b in input_data.to_batches(columns=["_index_", "category", "name", "text"]):
            for item in b.to_pylist():
                input_items[item["_index_"]] = item

        for b in output_data.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            batch = []
            for item in b.to_pylist():
                index = item["_index_"]
                replication = item["_replication_"]
                covariates = input_items[index]
                name = covariates["name"]
                category = covariates["category"]
                input_text = covariates["text"]
                output_text = item["responses"][0]["text"]
                if input_text in output_text:
                    output_text = output_text.replace(input_text, "")
                raw_score = self.score_pat.search(output_text)
                try:
                    assert raw_score
                    matchedText = raw_score.group(0)
                    score = int(matchedText.replace(" ", ""))
                    output = GetResumeScoreScoredItem(
                        _index_=index,
                        _replication_=replication,
                        name=name,
                        category=category,
                        matchedText=matchedText,
                        resumeScore=score,
                    ).dict()
                except:
                    output = GetResumeScoreScoredItem(
                        _index_=index,
                        _replication_=replication,
                        name=name,
                        category=category,
                        matchedText=None,  # or: "ERROR"
                        resumeScore=None,
                    ).dict()
                batch.append(output)

            yield pyarrow.RecordBatch.from_pylist(batch, schema=self.schema)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import absl.app
import absl.flags
from absl import logging

from ...api import timestamp
from ...api.data.sinks import GoogleStorageUploader
from ...api.data.sources import ZenodoRecord

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string("source", None, "The source to obtain the data from.")
absl.flags.mark_flag_as_required("source")

absl.flags.DEFINE_enum("source_kind", None, ["zenodo"], "The kind of source.")
absl.flags.mark_flag_as_required("source_kind")

absl.flags.DEFINE_string(
    "upload_bucket", None, "The name of the GS bucket to store the data in."
)
absl.flags.mark_flag_as_required("upload_bucket")

absl.flags.DEFINE_bool(
    "dryrun",
    False,
    "Set up the fetch operation but don't download data or modify anything.",
)


RETURNCODE_SUCCESS = 0
RETURNCODE_ERROR = 1


def _create_source():
    if FLAGS.source_kind == "zenodo":
        return ZenodoRecord(FLAGS.source)
    else:
        raise ValueError(f"Don't know how to fetch --source_kind={FLAGS.source_kind}")


def main(unused_argv):
    try:
        source = _create_source()
        sink = GoogleStorageUploader(FLAGS.upload_bucket)
        if not FLAGS.dryrun:
            ts_start = timestamp.now()
            logging.info(f"fetch started: {ts_start}")
            source.fetch_into(sink)
            ts_finish = timestamp.now()
            logging.info(f"fetch finished: {ts_finish}")
        return RETURNCODE_SUCCESS
    except:
        logging.exception("fetch may be incomplete")
        return RETURNCODE_ERROR


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import absl.app
from absl import logging

from .main import WorkflowsSink


def main(unused_argv):
    logging.set_verbosity(logging.DEBUG)

    app = WorkflowsSink()
    app.run()


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import abc
import shutil
import subprocess
from typing import Optional

import google.cloud.exceptions
import requests
from absl import logging
from google.cloud import storage

from dyff.schema.copydoc import copydoc

from ...api.exceptions import MissingDependencyError
from . import google_storage_md5


class DataSinkError(RuntimeError):
    pass


class DataUploadError(DataSinkError):
    pass


class DataSink:
    @property
    @abc.abstractmethod
    def sink_id(self) -> str:
        """The idenfitier of this data sink."""
        raise NotImplementedError()

    @abc.abstractmethod
    def resource_exists(self, sink_path: str) -> bool:
        """Checks if a valid resource exists at ``sink_path``.

        Args:
          sink_path: The identifier of the resource in the data sink.

        Returns:
          True if and only if a valid resource exists at ``sink_path``.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def upload_string(self, data: str, sink_path: str):
        """Uploads raw data in the form of a string.

        Args:
          data: The data to upload.
          sink_path: The identifier of the resource in the data sink.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def upload_from_http(
        self, source_url: str, sink_path: str, *, md5_hex: Optional[str] = None
    ):
        """Uploads the data at ``source_url`` to ``sink_path`` within the data
        sink.

        Args:
          source_url: The URL of the resource to upload.
          sink_path: The identifier of the resource in the data sink.
          md5_hex: If specified, the upload will fail and be rolled back if the
            md5 checksum of the uploaded data does not match this argument.
            Callers should provide this argument whenever possible.

        Raises:
          ConnectionError: If something goes wrong with network connections.
          DataUploadError: If the upload failed for some reason other than the
            connection.
        """
        raise NotImplementedError()


class GoogleStorageUploader(DataSink):
    def __init__(self, bucket_id: str):
        """Data sink that uploads to a Google Storage bucket.

        Args:
          bucket: A storage bucket that is already created.

        Raises:
          alignmentlabs.dyff.exceptions.MissingDependencyError: If necessary
            command-line tools are not installed.
          DataSinkError: If connecting to the bucket fails. This will often be
            due to authentication problems, or the bucket not existing.
        """
        if shutil.which("curl") is None:
            raise MissingDependencyError("'curl' is not installed")
        if shutil.which("gsutil") is None:
            raise MissingDependencyError("'gsutil' is not installed")

        client = storage.Client()
        # Note: get_bucket() requires more privileges than writing to a
        # specific bucket
        self._bucket_id = bucket_id
        self._bucket = client.bucket(bucket_id)
        try:
            self.resource_exists("_test")  # Just to see if it raises an exception
        except google.cloud.exceptions.Forbidden as ex:
            logging.error(f"Forbidden: {bucket_id}")
            raise DataSinkError(f"Forbidden: {bucket_id}") from ex
        except google.cloud.exceptions.NotFound as ex:
            logging.error(f"NotFound: {bucket_id}")
            raise DataSinkError(f"NotFound: {bucket_id}") from ex

    @property
    @copydoc(DataSink.sink_id)
    def sink_id(self) -> str:
        return f"gs://{self._bucket_id}"

    @copydoc(DataSink.resource_exists)
    def resource_exists(self, sink_path: str) -> bool:
        return self._bucket.get_blob(sink_path) is not None

    @copydoc(DataSink.upload_string)
    def upload_string(self, data: str, sink_path: str):
        blob = self._bucket.blob(sink_path)
        blob.upload_from_string(data, checksum="md5")

    @copydoc(DataSink.upload_from_http)
    def upload_from_http(
        self, source_url: str, sink_path: str, *, md5_hex: Optional[str] = None
    ):
        # I can't figure out how to check for errors from the 'curl' part of
        # the pipeline below, so we'll confirm the URL is working here.
        try:
            r = requests.head(source_url)
            r.raise_for_status()
        except requests.RequestException as ex:
            raise ConnectionError(f"something is wrong with {source_url}") from ex
        else:
            if r.status_code != 200:
                raise ConnectionError(
                    f"http error {r.status_code} (expected 200): {source_url}"
                )

        gsutil_args = ["gsutil"]
        if md5_hex is not None:
            md5_google = google_storage_md5(md5_hex=md5_hex)
            gsutil_args.extend(["-h", f"Content-MD5:{md5_google}"])
        gsutil_args.extend(["cp", "-", f"gs://{self._bucket.name}/{sink_path}"])
        gsutil = subprocess.Popen(
            gsutil_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        curl = subprocess.Popen(
            ["curl", source_url], stdout=gsutil.stdin, stderr=subprocess.DEVNULL
        )
        stdout, stderr = gsutil.communicate()

        if gsutil.returncode != 0:
            logging.error(stderr.decode())
            raise DataUploadError(f"upload failed with error code {gsutil.returncode}")


if __name__ == "__main__":
    client = storage.Client()
    bucket = client.get_bucket("alignmentlabs-rawdata-14a802b4b854421f")
    sink = GoogleStorageUploader(bucket)
    sink.upload_from_http("https://unsplash.com/photos/N_G2Sqdy9QY", "test/cat")
    print(sink.resource_exists("test/cat"))



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Optional

from dyff.schema.platform import Artifact, StorageSignedURL


class StorageBackend(abc.ABC):
    @abc.abstractmethod
    def storage_size(self, path: str) -> int:
        """Get the total size, in bytes, of all objects stored under the
        ``path`` prefix.
        """

    @abc.abstractmethod
    def list_dir(self, path: str) -> list[str]:
        """Get the absolute paths of all objects that are immediate
        "children" of ``path``.
        """

    @abc.abstractmethod
    def download_recursive(self, source: str, destination: str) -> None:
        """Download all objects stored under the path ``source`` to under the
        local directory ``destination``. The directory structure is preserved.
        """

    @abc.abstractmethod
    def upload_recursive(self, source: str, destination: str) -> None:
        """Upload all files stored under the local directory ``source`` to
        under the storage path ``destination``. The directory structure is
        preserved.
        """

    @abc.abstractmethod
    def signed_url_for_dataset_upload(
        self,
        dataset_id: str,
        artifact: Artifact,
        *,
        size_limit_bytes: Optional[int] = None,
        storage_path: Optional[str] = None,
    ) -> StorageSignedURL:
        """Create a temporary signed URL that can be used in a PUT request
        to upload an ``Artifact`` directly to storage.
        """

    @abc.abstractmethod
    def dataset_artifact_md5hash(
        self, dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
    ) -> bytes:
        """Compute the MD5 hash of a dataset artifact in storage."""



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import datetime
import json
from collections import defaultdict
from typing import Dict, Iterable, List, NamedTuple, Optional, TypeVar

import absl.app
import absl.flags
import confluent_kafka as kafka
import kubernetes.client.exceptions
import networkx
import pydantic
import rocksdict
from absl import logging

from dyff.api.backend.kafka.command import (
    deserialize_entity,
    deserialize_id,
    serialize_id,
    serialize_value,
)
from dyff.api.typing import YAMLObject
from dyff.core.config import config
from dyff.orchestrator.k8s import conditions, load_config, resources
from dyff.schema.platform import (
    Dataset,
    DyffEntity,
    DyffEntityType,
    Entities,
    EntityStatus,
    EntityStatusReason,
    is_status_success,
    is_status_terminal,
)

from . import crd

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string("namespace", "default", "k8s namespace")

absl.flags.DEFINE_bool("once", False, "Run the polling loop once, then exit")
absl.flags.DEFINE_bool(
    "from_beginning",
    False,
    "Consume the workflows.state topic from the beginning (i.e., reset the commit offset)",
)
absl.flags.DEFINE_bool(
    "disable_kafka_producer", False, "Print log messages but don't modify state"
)
absl.flags.DEFINE_bool(
    "disable_kubernetes_actions",
    False,
    "Don't make changes in Kubernetes. Still update the database and produce messages.",
)
absl.flags.DEFINE_bool("disable_quotas", False, "Don't enforce job quotas")

absl.flags.DEFINE_string("local_db_path", None, "Path to local DB files")

absl.flags.DEFINE_string("workflows_events_topic", None, "Workflows events topic")
absl.flags.DEFINE_string("workflows_state_topic", None, "Workflows state topic")
absl.flags.DEFINE_string("kafka_bootstrap_servers", None, "Address of bootstrap server")

absl.flags.DEFINE_bool(
    "debug_ignore_bad_messages",
    False,
    "Swallow exceptions from process_message(). Useful for testing where there are messages with obsolete schemas still in the topic.",
)
absl.flags.DEFINE_bool(
    "debug_ignore_missing_dependencies",
    False,
    "Assume that dependencies that are not in the dependency graph are in a Ready status.",
)


# FIXME: Should have per-account quotas stored in the database
quotas: dict[str, int] = {
    Entities.Audit: 8,
    Entities.Dataset: 1,
    Entities.Evaluation: 4,
    Entities.InferenceService: 4,
    Entities.InferenceSession: 4,
    Entities.Model: 4,
    Entities.Report: 8,
}


class LocalDBWriteBatch:
    """Represents a batch write operation.

    Apply the batch write with ``LocalDB.write_batch()``. The batch write happens
    in a single transaction.
    """

    def __init__(self):
        self.batch = rocksdict.WriteBatch(raw_mode=True)

    def put_entity(self, id: str, entity: DyffEntity):
        """Add an entity put operation to the batch."""
        self.batch.put(id.encode("utf-8"), entity.json().encode("utf-8"))

    def put_dependency_graph(self, dependency_graph: DependencyGraph):
        """Add a DependencyGraph put operation to the batch."""
        self.batch.put(
            "__dependency_graph".encode("utf-8"), dependency_graph.serialize()
        )


class LocalDB:
    """A local database that stores the Orchestrator's snapshot of the state of
    all the workflow entities in the system as well as the current dependency
    graph.
    """

    def __init__(self, db_path: str):
        self.db = rocksdict.Rdict(db_path, options=rocksdict.Options(raw_mode=True))

    def close(self):
        if self.db:
            self.db.close()

    def __getitem__(self, id: str) -> Optional[DyffEntityType]:
        return self.get_entity(id)

    def get_entity(self, id: str) -> Optional[DyffEntityType]:
        """Get an entity by ID."""
        raw = self.db.get(id.encode("utf-8"))
        return raw and deserialize_entity(raw)

    def get_dependency_graph(self) -> DependencyGraph:
        """Get the dependency graph."""
        graph_bytes: Optional[bytes] = self.db.get("__dependency_graph".encode("utf-8"))
        graph = DependencyGraph()
        if graph_bytes is None:
            # This doesn't need to be part of a batch transaction because the
            # graph is empty.
            self.db.put("__dependency_graph".encode("utf-8"), graph.serialize())
        else:
            graph.deserialize(graph_bytes)
        return graph

    def write_batch(self, batch: LocalDBWriteBatch):
        """Execute a batch write.

        The batch write happens in a single transaction, so it either totally
        succeeds or it fails without changing the DB state.
        """
        return self.db.write(batch.batch)


DyffEntityT = TypeVar("DyffEntityT", bound=DyffEntity)


class DependencyGraph:
    """A directed graph representing dependencies among workflows."""

    def __init__(self):
        self._graph = networkx.DiGraph()

    def add_created(self, db: LocalDB, entity: DyffEntityT) -> YAMLObject:
        """Add a newly-created workflow to the graph.

        Returns either the original entity unchanged, or a copy of the entity with
        its .status and .reason fields set appropriately. The returned entity
        may have .reason == UnsatisfiedDependency if it is waiting for a dependency
        to complete, or .status == Error and .reason == FailedDependency if one
        of its dependencies is already in a failure state.
        """
        assert entity.status == EntityStatus.created
        assert entity.reason == None
        edges = []
        update = {}
        for dependency_id in entity.dependencies():
            dependency = db.get_entity(dependency_id)
            if dependency is None:
                if not FLAGS.debug_ignore_missing_dependencies:
                    edges.append((dependency_id, entity.id))
                    update = {
                        "status": EntityStatus.created,
                        "reason": EntityStatusReason.unsatisfied_dependency,
                    }
            elif not is_status_terminal(dependency.status):
                edges.append((dependency_id, entity.id))
                update = {
                    "status": EntityStatus.created,
                    "reason": EntityStatusReason.unsatisfied_dependency,
                }
            elif not is_status_success(dependency.status):
                # We treat this as an unrecoverable error
                update = {
                    "status": EntityStatus.error,
                    "reason": EntityStatusReason.failed_dependency,
                }
                break
        else:
            # Delay graph changes until here in case there's a failed dependency
            self._graph.add_node(entity.id)
            self._graph.add_edges_from(edges)
        return update

    def on_workflow_termination(
        self, db: LocalDB, entity: DyffEntityType
    ) -> List[DyffEntityType]:
        """Update the graph when a workflow terminates.

        Returns a list of entities whose .status and/or .reason also changed as
        a result of the termination. The returned list *does not* include the
        ``entity`` argument.

        Currently, the only reason that other entities' state
        would change is if the workflow terminated in a failure state and other
        workflows depend on it.
        """

        def propagate_failure(root: DyffEntityType) -> List[DyffEntityType]:
            failures: List[DyffEntityType] = []
            dfs_stack: List[str] = [root.id]
            visited = set()
            while dfs_stack:
                current_node = dfs_stack.pop()
                if current_node in visited:
                    continue
                visited.add(current_node)
                dfs_stack.extend(self._graph.successors(current_node))
                if current_node != root.id:
                    # Don't update the status of the root entity, to preserve its .reason
                    current_entity = db.get_entity(current_node)
                    assert current_entity is not None
                    update = {
                        "status": EntityStatus.error,
                        "reason": EntityStatusReason.failed_dependency,
                    }
                    failures.append(current_entity.copy(update=update))

            for failed_entity in failures:
                self._graph.remove_node(failed_entity.id)
            return failures

        assert is_status_terminal(entity.status)
        if not self._graph.has_node(entity.id):
            # If we receive a duplicate message, we may have already removed the node
            return []

        failures = []
        if not is_status_success(entity.status):
            failures = propagate_failure(entity)
        self._graph.remove_node(entity.id)
        return failures

    def schedulable_entities(self) -> Iterable[str]:
        """Yields the .id's of entities in the graph that have all of their
        dependencies satisfied.
        """
        yield from (id for id, in_degree in self._graph.in_degree() if in_degree == 0)

    def deserialize(self, data: bytes):
        """Restore the graph structure from its serialized representation."""
        graph_data = json.loads(data.decode("utf-8"))
        self._graph = networkx.readwrite.json_graph.adjacency_graph(
            graph_data, directed=True, multigraph=False
        )

    def serialize(self) -> bytes:
        """Returns the serialized representation of the graph structure."""
        graph_data = networkx.readwrite.json_graph.adjacency_data(self._graph)
        return json.dumps(graph_data).encode("utf-8")

    def __str__(self) -> str:
        graph_data = networkx.readwrite.json_graph.adjacency_data(self._graph)
        return json.dumps(graph_data, indent=2)


class JobQueueItem(NamedTuple):
    # This should be 0 for admitted entities and 1 otherwise, so that admitted
    # entities are sorted before not-admitted entities.
    waiting: bool
    creationTime: datetime.datetime
    # .id is included as a tiebreaker in case .creationTime's are equal, since
    # DyffEntity's don't have an ordering defined
    id: str
    entity: DyffEntityType

    @staticmethod
    def from_entity(entity: DyffEntityType) -> JobQueueItem:
        return JobQueueItem(
            waiting=(entity.status != EntityStatus.admitted),
            creationTime=entity.creationTime,
            id=entity.id,
            entity=entity,
        )


class Orchestrator:
    """Manages Dyff entities that are in the ``Created`` status until advancing
    them to the ``Admitted`` status.

    The Orchestrator maintains a local snapshot of entity state in a RocksDB
    instance. Incoming Kafka messages update the local snapshot. Periodically,
    the Orchestrator runs a scheduling task using its local snapshot of entity
    state.

    The Orchestrator also maintains a dependency graph, which is also stored
    in RocksDB. This records which workflows are waiting on dependencies to
    complete.
    """

    def __init__(
        self,
        *,
        local_db_path: str,
        workflows_events_topic: str,
        workflows_state_topic: str,
        kafka_bootstrap_servers: str,
        disable_kafka_producer: bool = False,
        disable_kubernetes_actions: bool = False,
    ):
        self.workflows_events_topic = workflows_events_topic
        self.workflows_state_topic = workflows_state_topic

        self.db = LocalDB(local_db_path)
        self.dependency_graph = self.db.get_dependency_graph()
        self.disable_kafka_producer = disable_kafka_producer
        self.disable_kubernetes_actions = disable_kubernetes_actions

        consumer_config = {
            "bootstrap.servers": kafka_bootstrap_servers,
            "group.id": "dyff.system.orchestrator",
            "auto.offset.reset": "earliest",
            "enable.auto.commit": False,
            "on_commit": self.on_commit,
        }
        self.consumer = kafka.Consumer(consumer_config)

        producer_config = {
            "bootstrap.servers": kafka_bootstrap_servers,
            "client.id": "dyff.system.orchestrator",
            # This is the default, but setting it explicitly means an exception will
            # be raised if other config settings are in conflict.
            "enable.idempotence": True,
        }
        self.producer = kafka.Producer(producer_config)

        self.log_debug_db()

    def log_debug_db(self):
        if logging.level_debug():
            db_lines = []
            for k, v in self.db.db.items():
                db_lines.append(f"{k}: {v}")
            db_str = "\n".join(db_lines)
            logging.debug(f"LocalDB:\n{db_str}")
            logging.debug(f"DependencyGraph:\n{self.dependency_graph}")

    def close(self):
        if self.producer:
            self.producer.flush()
        if self.consumer:
            self.consumer.close()
        if self.db:
            self.db.close()

    def _admit_entity(self, entity: DyffEntityT, manifest) -> YAMLObject:
        assert entity.status == EntityStatus.created

        conditions_list = conditions.ConditionsList()
        conditions_list.set(
            EntityStatus.admitted, status=conditions.ConditionStatus.true
        )
        manifest["status"] = {"conditions": conditions_list.conditions_list}

        if not self.disable_kubernetes_actions:
            resources.create_resource(entity.kind, manifest, namespace=FLAGS.namespace)
        logging.info(f"created {manifest['metadata']['name']}")
        return {"status": EntityStatus.admitted, "reason": None}

    def on_produce_acknowledged(self, error, msg):
        if error:
            logging.error(f"delivery failed: {error}")
        else:
            logging.debug(f"message produced: {msg}")

    def on_commit(self, error, partitions):
        if error:
            logging.error(f"commit failed: {error}")
        else:
            logging.debug(f"committed partition offsets: {partitions}")

    def create_k8s_workflow(self, entity: DyffEntityType) -> Optional[YAMLObject]:
        try:
            manifest = crd.manifest(entity)
            if manifest is not None:
                return self._admit_entity(entity, manifest)
            else:
                logging.debug(f"Nothing to do for {entity.kind} {entity.id}")
                if isinstance(entity, Dataset):
                    # No status change: user needs to upload data to and then
                    # manually indicate Ready
                    return None
                else:
                    return {"status": EntityStatus.ready, "reason": None}
        except Exception:
            logging.exception(
                f"workflow {entity.kind} {entity.id}: failed to create k8s manifest"
            )
            # TODO: Make an enum entry for SchemaError reason (?)
            return {"status": EntityStatus.failed, "reason": "SchemaError"}

    def delete_k8s_workflow(self, entity: DyffEntity):
        name = resources.object_name(entity.kind, entity.id)
        if not self.disable_kubernetes_actions:
            resources.delete_resource(entity.kind, name, namespace=FLAGS.namespace)
        logging.info(f"deleted {name}")

    def produce_event(self, entity_id: str, update: YAMLObject):
        logging.debug(f"produce: {entity_id}: {update}")
        if not self.disable_kafka_producer:
            self.producer.produce(
                self.workflows_events_topic,
                key=serialize_id(entity_id),
                value=serialize_value(update),
                callback=self.on_produce_acknowledged,
            )

    def maybe_admit_more_workflows(self) -> None:
        """Scan the entire dependency graph for workflows whose dependencies are
        satisfied, and either admit them or place them in the .reason == QuotaLimit
        state as appropriate.

        Note that we do not modify the local db here; instead we send events
        and let the updates happen via the event processing logic. This is
        because events might change the dependency graph and we want to keep
        those updates in one place for simplicity.
        """
        # account -> kind -> [JobQueueItem]
        schedulable_entities: Dict[str, Dict[str, List[JobQueueItem]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for schedulable_id in self.dependency_graph.schedulable_entities():
            schedulable = self.db.get_entity(schedulable_id)
            if schedulable is not None:
                # Entity can be None if it is mentioned as a dependency but we haven't
                # received the Created message yet
                schedulable_entities[schedulable.account][schedulable.kind].append(
                    JobQueueItem.from_entity(schedulable)
                )

        for account, kinds in schedulable_entities.items():
            for kind, entities in kinds.items():
                # See JobQueueItem for ordering:
                #   * .status == Admitted items come before .status == Created items
                #   * Items with same status ordered by .creationTime
                entities.sort()
                for i, item in enumerate(entities):
                    if not item.waiting:
                        continue

                    if i < quotas[kind] or FLAGS.disable_quotas:
                        # TODO: per-account quotas

                        # Kubernetes is idempotent to multiple delivery because the jobs
                        # will have the same name and k8s will refuse to create the
                        # duplicate jobs, so k8s effectively turns at-least-once execution
                        # of the scheduling action into exactly-once execution.
                        try:
                            update = self.create_k8s_workflow(item.entity)
                        except Exception:
                            logging.exception("admission failed")
                            # TODO: Add enum entry for AdmissionFailed
                            update = {
                                "status": EntityStatus.error,
                                "reason": "AdmissionFailed",
                            }
                            if not FLAGS.debug_ignore_bad_messages:
                                raise
                        if update is not None:
                            # Failure here =>
                            #   * Next scheduling pass sees schedulable entity
                            #   * Double resource creation fails due to duplicate job name
                            #   => OK
                            self.produce_event(item.entity.id, update)
                            # Failure here =>
                            #   * Next scheduling pass sees schedulable entity
                            #   * Double resource creation fails due to duplicate job name
                            #   * Double event produced but Orchestrator will ignore it
                            #   => OK
                    elif item.entity.reason != EntityStatusReason.quota_limit:
                        # QuotaLimit
                        # The if-condition just reduces the number of redundant events
                        update = {
                            "status": EntityStatus.created,
                            "reason": EntityStatusReason.quota_limit,
                        }
                        self.produce_event(item.entity.id, update)
                        # Failure here =>
                        #   * Next scheduling pass sees blocked entity
                        #   * Double event produced but Orchestrator will ignore it
                        #   => OK

    def process_message(self, msg: kafka.Message):
        """Update the local RocksDB in response to an incoming Kafka message.

        The state of the entity in the message is always updated to keep the local
        snapshot in sync with the Kafka topic.

        The dependency graph is also updated in the following two cases:
          * A brand new entity in Created status (i.e., .reason is None)
          * An entity in a terminal status (finished or failed)

        No actions are taken here that modify the state of the system outside of
        the Orchestrator. Scheduling actions occur separately in
        ``maybe_admit_more_workflows()``, which runs periodically on a timer. The
        admission process may require significant computation if there are a lot
        of pending jobs, so executing it on every event would be excessive.
        """

        # To ensure fault tolerance, we need to make scheduling decisions based on
        # the current state, not on state transitions (similar to how the k8s
        # operator is implemented).
        #
        # Recording the state in the local DB is semantically similar to commiting
        # the consumer position in Kafka, i.e., it needs to happen *after* taking
        # any other actions so we have at-least-once execution of those actions.

        id = deserialize_id(msg.key())
        entity = deserialize_entity(msg.value())
        if id != entity.id:
            logging.error(
                f"message key {id} != entity.id {entity.id}; skipping message"
            )
            return
        logging.info(f"process_message(): {id}: {entity}")

        # Short-circuit for some duplicate messages that are easy to detect.
        # This doesn't catch all double message deliveries, because failure between
        # db.put() and consumer.commit() leaves db ahead of consumer.
        stored_entity = self.db.get_entity(entity.id)
        if stored_entity == entity:
            logging.debug(f"rejected duplicate message: {entity}")
            return

        # RocksDB batch write happens in a single transaction
        write_batch = LocalDBWriteBatch()

        # These are the two status events that might change the dependency graph
        if entity.status == EntityStatus.created and entity.reason is None:
            update = self.dependency_graph.add_created(self.db, entity)
            write_batch.put_dependency_graph(self.dependency_graph)
            if update:
                self.produce_event(entity.id, update)
        elif is_status_terminal(entity.status):
            if entity.status in [EntityStatus.deleted, EntityStatus.terminated]:
                # TODO: We could delete all k8s entities that are in terminal
                # states, but it's useful to keep them around for debugging.
                # Alternatively, we could pass the Terminated status through
                # to the operator, so it can delete child workflows but leave
                # the parent workflow in k8s for examination.
                try:
                    self.delete_k8s_workflow(entity)
                except kubernetes.client.exceptions.ApiException as ex:
                    # NotFound is OK because we might have a duplicate message
                    # or a user might delete the entity after the k8s resource
                    # has been GC'd.
                    if ex.status != 404:
                        raise
            changed_dependents: List[DyffEntityType] = (
                self.dependency_graph.on_workflow_termination(self.db, entity)
            )
            for e in changed_dependents:
                self.produce_event(e.id, {"status": e.status, "reason": e.reason})
                write_batch.put_entity(e.id, e)
            write_batch.put_dependency_graph(self.dependency_graph)

        # Store state change in local DB
        write_batch.put_entity(entity.id, entity)
        # DB write is last step so that we get at-least-once execution of the
        # dependency graph update.
        self.db.write_batch(write_batch)
        # Failure here =>
        #   * State msg received again
        #   * (Possible) Event msg produced again but orchestrator will ignore
        #   * Double put in DB but the value is the same
        #   => OK
        self.log_debug_db()

    def run(self, *, once: bool = False):
        """Run the Orchestrator loop forever.

        If ``once == True``, return after running the loop once, including one run
        of the scheduling step.
        """

        def datetime_now() -> datetime.datetime:
            return datetime.datetime.now(datetime.timezone.utc)

        def reset_partition_offsets_to_beginning(consumer, partitions):
            for partition in partitions:
                # Note: -2 is the Kafka special value meaning "from beginning"
                partition.offset = -2
            consumer.assign(partitions)

        scheduler_interval = datetime.timedelta(seconds=10)
        subscriptions = [self.workflows_state_topic]

        if FLAGS.from_beginning:
            self.consumer.subscribe(
                subscriptions, on_assign=reset_partition_offsets_to_beginning
            )
        else:
            self.consumer.subscribe(subscriptions)
        # This starts as None so that scheduling always runs on the first iteration,
        # otherwise the --once flag would be fairly useless
        then = None
        while True:
            # Kafka messages update the local DB state, but we only act on them
            # periodically
            msg = self.consumer.poll(timeout=1.0)
            if msg is not None:
                if msg.error():
                    if msg.error().code() == kafka.KafkaError._PARTITION_EOF:
                        logging.warning(
                            f"{msg.topic()}:{msg.partition()} EOF at offset {msg.offset()}"
                        )
                    else:
                        raise kafka.KafkaException(msg.error())
                else:
                    try:
                        # Want "at least once" delivery of state message
                        #   => "process" must happen before "commit"
                        self.process_message(msg)
                    except Exception:
                        logging.error(f"bad message: {msg.value()}")
                        if not FLAGS.debug_ignore_bad_messages:
                            raise
                    # Failure here =>
                    #   * DB update happens again
                    #   => OK (see process_message)
                    self.consumer.commit(asynchronous=True)

            # Periodically check if we can schedule more stuff
            now = datetime_now()
            if (then is None) or (now - then >= scheduler_interval):
                try:
                    self.maybe_admit_more_workflows()
                except pydantic.ValidationError:
                    logging.error(f"schema error in DB")
                    if not FLAGS.debug_ignore_bad_messages:
                        raise
                then = now

            if once:
                break


def main(_unused_argv):
    load_config()

    orchestrator = Orchestrator(
        local_db_path=FLAGS.local_db_path,
        workflows_events_topic=(
            FLAGS.workflows_events_topic or config.kafka.topics.workflows_events
        ),
        workflows_state_topic=(
            FLAGS.workflows_state_topic or config.kafka.topics.workflows_state
        ),
        kafka_bootstrap_servers=(
            FLAGS.kafka_bootstrap_servers or config.kafka.config.bootstrap_servers
        ),
        disable_kafka_producer=FLAGS.disable_kafka_producer,
        disable_kubernetes_actions=FLAGS.disable_kubernetes_actions,
    )
    try:
        orchestrator.run(once=FLAGS.once)
    finally:
        orchestrator.close()



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""This is where we translate from the Dyff API schema to the Dyff k8s operator
schema.

A primary design goal for the k8s operator is that it should be useful
without the rest of the Dyff system. Meaning, for example,
* The operator can't look up other resources by reference, since it doesn't
  have a database.
* We want the k8s API to be more stable than the Dyff API, so it has to allow
  for more customization.

The two schemas share a lot of common structure, but as things have grown
more complex they have needed to diverge to prevent excessive coupling between
the k8s operator and the rest of the system.
"""

from __future__ import annotations

import json
from typing import Any, Optional
from urllib.parse import urlparse

from dyff.core.config import config
from dyff.orchestrator.k8s import resources
from dyff.schema.platform import (
    Dataset,
    DyffEntityType,
    Evaluation,
    InferenceService,
    InferenceServiceRunnerKind,
    InferenceSession,
    InferenceSessionSpec,
    Model,
    ModelArtifactKind,
    ModelStorageMedium,
    Report,
    SchemaAdapter,
)


def manifest(entity: DyffEntityType) -> Optional[dict]:
    if isinstance(entity, Dataset):
        # Currently not supporting a "fetch_data" operation
        return None
    elif isinstance(entity, Evaluation):
        return evaluation_manifest(entity)
    elif isinstance(entity, InferenceService):
        return inference_service_manifest(entity)
    elif isinstance(entity, InferenceSession):
        return inference_session_manifest(entity)
    elif isinstance(entity, Model):
        return model_manifest(entity)
    elif isinstance(entity, Report):
        return report_manifest(entity)
    else:
        raise TypeError(f"entity kind: {entity.kind}")


def model_manifest(model: Model) -> Optional[dict]:
    """Translate a Dyff API ``Model`` to a k8s ``Model``."""
    spec: dict[str, Any] = {
        "id": model.id,
        "account": model.account,
        # Note: We're assuming this sub-schema matches k8s exactly
        "source": model.source.dict(),
    }

    if model.storage.medium == ModelStorageMedium.PersistentVolume:
        spec["storage"] = {
            "kind": "PersistentVolume",
            "persistentVolume": {"storageClassName": "dyff-model"},
        }
    elif model.storage.medium == ModelStorageMedium.ObjectStorage:
        url = urlparse(config.resources.models.storage.url)
        protocol = url.scheme
        bucketName = url.hostname
        spec["storage"] = {
            "kind": "ObjectStore",
            "objectStore": {
                "protocol": protocol,
                "bucketName": bucketName,
            },
        }
    else:
        raise NotImplementedError(f"model.storage.medium: {model.storage.medium}")

    spec["storage"]["quantity"] = model.resources.storage

    return resources.model_manifest(spec)


def inference_service_manifest(inference_service: InferenceService) -> Optional[dict]:
    """Translate a Dyff API ``InferenceService`` to a k8s ``InferenceService``."""
    if inference_service.builder is None:
        # Nothing to do
        return None

    if inference_service.builder.kind == "bentoml-transformers-pipeline":
        # TODO: This is the original build flow for huggingface Pipeline models
        raise NotImplementedError()

    return None


def _inference_session_template_spec(
    session: InferenceSessionSpec,
) -> dict:
    spec: dict[str, Any] = {
        "replicas": session.replicas,
        "useSpotPods": session.useSpotPods,
    }
    # TODO: Define this as a constant somewhere
    model_mount_path = "/dyff/mnt/model"

    service = session.inferenceService
    if service.runner is None:
        raise ValueError(f"InferenceService {service.id} has no .runner")
    if service.model is not None and service.model.storage.medium not in [
        None,
        ModelStorageMedium.PersistentVolume,
    ]:
        raise NotImplementedError(
            f"service.model.storage.medium: {service.model.storage.medium}"
        )

    if service.runner.kind == InferenceServiceRunnerKind.STANDALONE:
        spec["image"] = config.orchestrator.images.standalone.format(service=service)
    elif service.runner.kind == InferenceServiceRunnerKind.MOCK:
        spec["image"] = config.orchestrator.images.mock
    elif service.runner.kind == InferenceServiceRunnerKind.BENTOML_SERVICE_OPENLLM:
        # TODO: Probably just remove this Kind
        raise NotImplementedError()
    elif service.runner.kind in [
        InferenceServiceRunnerKind.HUGGINGFACE,
        InferenceServiceRunnerKind.VLLM,
    ]:
        # Runners for HuggingFace models
        if service.model is None:
            raise ValueError(f"runner {service.runner.kind}: service.model is required")
        if service.model.artifact.kind != ModelArtifactKind.HuggingFaceCache:
            raise ValueError(
                f"runner {service.runner.kind}:"
                f" service.model.artifact.kind must be {ModelArtifactKind.HuggingFaceCache}"
            )
        hf_cache_artifact = service.model.artifact.huggingFaceCache
        assert hf_cache_artifact is not None

        if service.runner.kind == InferenceServiceRunnerKind.HUGGINGFACE:
            spec["image"] = config.orchestrator.images.huggingface
            spec["args"] = [
                "--model_name",
                hf_cache_artifact.repoID,
                "--model_revision",
                hf_cache_artifact.revision,
            ]
        elif service.runner.kind == InferenceServiceRunnerKind.VLLM:
            spec["image"] = config.orchestrator.images.vllm
            spec["args"] = [
                "--model",
                f"{model_mount_path}/{hf_cache_artifact.snapshot_path()}",
                "--download-dir",
                model_mount_path,
            ]
        else:
            raise AssertionError(f"Unexpected runner: {service.runner.kind}")

        if service.runner.args:
            spec["args"].extend(service.runner.args)
        # Some of these are allegedly redundant, but I don't trust HF not
        # to break things
        spec["env"] = [
            {"name": "HF_DATASETS_OFFLINE", "value": "1"},
            {"name": "HF_HOME", "value": model_mount_path},
            {"name": "HUGGINGFACE_HUB_CACHE", "value": model_mount_path},
            {"name": "TRANSFORMERS_CACHE", "value": model_mount_path},
            {"name": "TRANSFORMERS_OFFLINE", "value": "1"},
            {"name": "HF_MODULES_CACHE", "value": "/tmp/hf_modules"},
        ]
        spec["dependencies"] = [
            {
                "kind": "ReadOnlyVolume",
                "readOnlyVolume": {
                    "name": "model",
                    "claimName": f"model-{service.model.id}-rox",
                    "mountPath": model_mount_path,
                },
            }
        ]
    else:
        raise NotImplementedError(f"service.runner.kind {service.runner.kind}")

    session_resources = {
        # defaults
        "requests": {
            "memory": "4Gi",
        }
    }
    if service.runner.resources is not None:
        session_resources["requests"].update(
            {k: v for k, v in service.runner.resources if k in ["cpu", "memory"]}
        )
    spec["resources"] = session_resources

    if session.accelerator is not None:
        spec["accelerator"] = session.accelerator.dict()
    elif service.runner.accelerator is not None:
        spec["accelerator"] = service.runner.accelerator.dict()

    return spec


def inference_session_manifest(inference_session: InferenceSession) -> Optional[dict]:
    """Translate a Dyff API ``InferenceSession`` to a k8s ``InferenceSession``."""
    spec: dict[str, Any] = {
        "id": inference_session.id,
        "account": inference_session.account,
    }
    template_spec = _inference_session_template_spec(inference_session)
    spec.update(template_spec)
    return resources.inference_session_manifest(spec)


def evaluation_manifest(evaluation: Evaluation) -> Optional[dict]:
    """Translate a Dyff API ``Evaluation`` to a k8s ``Evaluation``."""
    # Adapters take arbitrary json configuration. k8s can't handle this, so
    # we encode it as a string.
    interface_spec = evaluation.inferenceSession.inferenceService.interface.dict()
    if (inputPipeline := interface_spec.get("inputPipeline")) is not None:
        for adapter in inputPipeline:
            adapter["configuration"] = json.dumps(adapter["configuration"])
    if (outputPipeline := interface_spec.get("outputPipeline")) is not None:
        for adapter in outputPipeline:
            adapter["configuration"] = json.dumps(adapter["configuration"])
    spec: dict[str, Any] = {
        "id": evaluation.id,
        "account": evaluation.account,
        "dataset": evaluation.dataset,
        "inferenceSession": _inference_session_template_spec(
            evaluation.inferenceSession
        ),
        "interface": interface_spec,
        "replications": evaluation.replications,
        "workersPerReplica": evaluation.workersPerReplica,
    }
    return resources.evaluation_manifest(spec)


def report_manifest(report: Report) -> Optional[dict]:
    """Translate a Dyff API ``Report`` to a k8s ``Report``."""

    def pipeline_configuration(adapters: list[SchemaAdapter]):
        pipeline_json = []
        for adapter in adapters:
            pipeline_json.append(
                {
                    "kind": adapter.kind,
                    "configuration": json.dumps(adapter.configuration),
                }
            )
        return pipeline_json

    spec = {
        "id": report.id,
        "account": report.account,
        "report": report.rubric,
        "dataset": report.dataset,
        "evaluation": report.evaluation,
    }
    if report.datasetView is not None:
        if report.datasetView.adapterPipeline:
            spec["datasetAdapter"] = pipeline_configuration(
                report.datasetView.adapterPipeline
            )
    if report.evaluationView is not None:
        if report.evaluationView.adapterPipeline:
            spec["evaluationAdapter"] = pipeline_configuration(
                report.evaluationView.adapterPipeline
            )
    return resources.report_manifest(spec)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from dyff.client import Client
from dyff.schema.requests import ReportCreateRequest

API_KEY: str = ...
ACCOUNT: str = ...

dyffapi = Client(api_key=API_KEY)

report_request = ReportCreateRequest(
    account=ACCOUNT,
    rubric="example.LikeCount",
    evaluation="825e3b76d74a446e997a8ea9c78b2d42",
)

report = dyffapi.reports.create(report_request)
print(f"created report:\n{report}")



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

from typing import NamedTuple, Sequence

import absl.app
import absl.flags
from absl import logging

from dyff.api import storage

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_multi_string(
    "transfer",
    None,
    "A string like 'gs://source/foo>/local/bar'."
    " Note the '>' character separating source and destination."
    " The *entire directory* 'foo' will end up at '/local/bar/foo'."
    " If source is a storage path and destination is a filesystem path,"
    " download from storage."
    " If source is a local path and destination is a storage path,"
    " upload to storage.",
)
absl.flags.mark_flags_as_required(["transfer"])


class ResourceLocation(NamedTuple):
    path: str
    protocol: str

    def uri(self) -> str:
        if self.protocol == "filesystem":
            return self.path
        else:
            return f"{self.protocol}://{self.path}"

    @staticmethod
    def from_path(path: str) -> ResourceLocation:
        parts = path.split("://")
        if len(parts) == 1:
            return ResourceLocation(path=parts[0], protocol="filesystem")
        elif len(parts) == 2:
            return ResourceLocation(path=parts[1], protocol=parts[0])
        else:
            raise ValueError(
                "expected filesystem path (/some/path) or storage path (gs://bucket/path)"
                f"; got '{path}'"
            )


class Transfer(NamedTuple):
    source: ResourceLocation
    destination: ResourceLocation


def main(_argv: Sequence[str]):
    supported_protocols = ["filesystem", "gs"]
    # Validate everything first
    transfers: list[Transfer] = []
    for command in FLAGS.transfer:
        source_path, destination_path = str(command).split(">")
        transfer = Transfer(
            source=ResourceLocation.from_path(source_path.strip()),
            destination=ResourceLocation.from_path(destination_path.strip()),
        )
        if not (
            (transfer.source.protocol == "filesystem")
            ^ (transfer.destination.protocol == "filesystem")
        ):
            raise NotImplementedError(
                "only storage>filesystem and filesystem>storage are implemented"
            )
        if transfer.source.protocol not in supported_protocols:
            raise NotImplementedError(
                f"unsupported protocol {transfer.source.protocol}"
            )
        if transfer.destination.protocol not in supported_protocols:
            raise NotImplementedError(
                f"unsupported protocol {transfer.destination.protocol}"
            )
        transfers.append(transfer)

    for transfer in transfers:
        if transfer.source.protocol != "filesystem":
            logging.info(
                f"download: {transfer.source.uri()} > {transfer.destination.uri()}"
            )
            storage.download_recursive(
                transfer.source.uri(), transfer.destination.uri()
            )
        elif transfer.destination.protocol != "filesystem":
            logging.info(
                f"upload: {transfer.source.uri()} > {transfer.destination.uri()}"
            )
            storage.upload_recursive(transfer.source.uri(), transfer.destination.uri())
        else:
            raise AssertionError()


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import base64
import hashlib
from typing import Dict, List

import pyarrow

from ..typing import PathLike


def parse_dataset_filters(filters: List[Dict]):
    filter_expression = None
    for f in filters:
        field = pyarrow.compute.field(f["field"])
        value = f["value"]
        relation = f["relation"]
        if relation == "==":
            expr = field == value
        elif relation == "!=":
            expr = field != value
        else:
            raise NotImplementedError()

        if filter_expression is None:
            filter_expression = expr
        else:
            filter_expression &= expr
    return filter_expression


def content_hash(content: bytes) -> str:
    h = hashlib.sha256()
    h.update(content)
    return h.hexdigest()


def google_storage_md5(
    *, data: bytes = None, file: PathLike = None, md5_hex: str = None
) -> str:
    if sum(int(x is not None) for x in [data, file, md5_hex]) != 1:
        raise ValueError("must specify exactly 1 argument")

    h = hashlib.md5()
    if data:
        h.update(data)
        digest = h.digest()
    elif file:
        with open(file, "rb") as fin:
            h.update(fin.read())
        digest = h.digest()
    else:
        digest = bytes.fromhex(md5_hex)
    return base64.b64encode(digest).decode()


def file_checksum(file: PathLike, *, algorithm="md5", format="hex") -> str:
    if algorithm == "md5":
        h = hashlib.md5()
    else:
        raise ValueError(f"Unsupported algorithm: {algorithm}")

    with open(file, "rb") as fin:
        h.update(fin.read())

    if format == "hex":
        return h.hexdigest()
    elif format == "base64":
        return base64.b64encode(h.digest()).decode()
    else:
        raise ValueError(f"Unsupported format: {format}")


def verify_file_checksum(file: PathLike, checksum: bytes) -> bool:
    algorithm, digest = checksum.split(":")
    return digest == file_checksum(file, algorithm=algorithm)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""Metrics are summaries of the scores generated from Rubrics.
"""

from . import text
from .base import Metric

__all__ = ["Metric", "text"]



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import json

import click
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from dyff.schema.platform import AccessGrant, APIFunctions, Resources


def admin(account_id: str) -> dict:
    grants = [
        AccessGrant(
            resources=[Resources.ALL],
            functions=[APIFunctions.all],
            accounts=["*"],
            entities=["*"],
        ),
    ]
    return {
        "account": account_id,
        "grants": grants,
    }


def audit_developer(account_id: str) -> dict:
    grants = [
        AccessGrant(
            resources=[Resources.ALL],
            functions=[APIFunctions.get, APIFunctions.query],
            accounts=["*"],
            entities=["*"],
        ),
        AccessGrant(
            resources=[Resources.ALL],
            functions=[APIFunctions.consume, APIFunctions.data],
            accounts=["public"],
        ),
        AccessGrant(
            resources=[Resources.ALL],
            functions=[APIFunctions.all],
            accounts=[account_id],
        ),
    ]
    return {
        "account": account_id,
        "grants": grants,
    }


def named_role(role_name: str, account_id: str) -> dict:
    if role_name == "Admin":
        return admin(account_id)
    elif role_name == "AuditDeveloper":
        return audit_developer(account_id)
    else:
        raise ValueError(f"Unknown role: {role_name}")


@click.group()
def roles():
    pass


@roles.command()
@click.argument("role", type=str, metavar="ROLE", required=True)
@click.option(
    "account_id", "--id", "-i", type=str, metavar="ID", required=True, help="Account ID"
)
def get(role: str, account_id: str):
    try:
        config = named_role(role, account_id)
        # Easy way to convert enum members to strings
        config["grants"] = [json.loads(grant.json()) for grant in config["grants"]]
        s = StringIO()
        YAML().dump(config, s)
        click.echo(s.getvalue())
    except Exception as ex:
        raise click.ClickException(" ".join(ex.args)) from ex



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""Tools for working with input and output data for systems being audited.
"""

from . import text

__all__ = [
    "text",
]



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import absl.app
import absl.flags
from absl import logging

from dyff.audit.scoring import Rubric
from dyff.core import dynamic_import
from dyff.schema.dataset import arrow

# -----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "rubric", None, "Qualified Python module name of the Rubric to apply"
)

absl.flags.DEFINE_string("dataset_path", None, "Local path to the Dataset data")

absl.flags.DEFINE_string("evaluation_path", None, "Local path to the Evaluation data")

absl.flags.DEFINE_string(
    "output_path", None, "Local path to where the report output should go"
)

absl.flags.mark_flags_as_required(
    ["rubric", "dataset_path", "evaluation_path", "output_path"]
)

# -----------------------------------------------------------------------------


def main(_unused_argv) -> None:
    logging.set_verbosity(logging.INFO)

    if FLAGS.rubric.startswith("dyff."):
        # Don't accept arbitrary import paths
        rubric = dynamic_import.instantiate(FLAGS.rubric)
    else:
        # Backwards compatibility
        rubric = dynamic_import.instantiate(f"dyff.audit.scoring.{FLAGS.rubric}")
    if not isinstance(rubric, Rubric):
        raise ValueError(f"{FLAGS.rubric} is not a Rubric subclass")

    task_data = arrow.open_dataset(FLAGS.dataset_path)
    outputs_data = arrow.open_dataset(FLAGS.evaluation_path)

    # TODO: We're not doing anything with the DataViews yet

    arrow.write_dataset(
        rubric.apply(task_data, outputs_data),
        output_path=FLAGS.output_path,
        feature_schema=rubric.schema,
    )


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime, timedelta

from dyff.client import Client
from dyff.schema.requests import (
    EvaluationCreateRequest,
    EvaluationInferenceSessionRequest,
)

API_KEY: str = ...
ACCOUNT: str = ...

dyffapi = Client(api_key=API_KEY)

# databricks/dolly-v2-3b
service_id = "de449a3995b54f13be8c64f876131032"
# linux-fortune
dataset_id = "0e702977e2864cbd8d99654c7a0d80a5"

evaluation_request = EvaluationCreateRequest(
    account=ACCOUNT,
    dataset=dataset_id,
    inferenceSession=EvaluationInferenceSessionRequest(
        inferenceService=service_id,
        expires=datetime.utcnow() + timedelta(days=1),
        replicas=1,
        useSpotPods=True,
    ),
    # There is a pool of replicas*workersPerReplia client connections that
    # send requests to the inference session. vLLM collects individual requests
    # into adaptively-sized batches. Generally, you would want to choose this
    # value by increasing it until you stop getting higher throughput.
    workersPerReplica=32,
)

evaluation = dyffapi.evaluations.create(evaluation_request)
print(evaluation)



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import secrets

import absl.app
import absl.flags
import google.cloud.exceptions
import google.cloud.storage

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "category",
    None,
    "Create an s3 bucket named named 'alignmentlabs-<category>-<random ID>'",
)
absl.flags.mark_flag_as_required("category")

absl.flags.DEFINE_integer(
    "conflict_retries",
    5,
    "Number of times to re-try with different random ID if there is a name conflict",
)


RETURNCODE_SUCCESS = 0
RETURNCODE_CONFLICTS = 1


def _next_bucket_name(category):
    hex = secrets.token_hex(8)
    return f"alignmentlabs-{category}-{hex}"


def main(unused_argv):
    client = google.cloud.storage.Client()

    for i in range(FLAGS.conflict_retries):
        bucket_name = _next_bucket_name(FLAGS.category)
        try:
            new_bucket = client.create_bucket(bucket_name, location="us-central1")
        except google.cloud.exceptions.Conflict:
            continue
        else:
            print(f"gs://{bucket_name}")
            return RETURNCODE_SUCCESS
    else:
        print("E: too many name conflicts")
        return RETURNCODE_CONFLICTS


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""Testing version of the Web server for the dyff platform.
"""

import os
import warnings

import fastapi

from dyff.schema.platform import APIKey

from ..web import routes as web_routes
from . import api, tokens


def _not_kubernetes() -> None:
    if os.environ.get("KUBERNETES_SERVICE_HOST"):
        raise fastapi.HTTPException(
            fastapi.status.HTTP_503_SERVICE_UNAVAILABLE,
            detail="Test configuration is not to be deployed!",
        )


class NonVerifyingJWTSigner(tokens.JWTSigner):
    def __init__(self):
        super().__init__(None)

    def unsign_api_key(self, token: str) -> APIKey:
        warnings.warn("Not verifying JWT signature")
        return super()._unsign_api_key(
            token, decode_options={"verify_signature": False}
        )


api.get_api_key_signer = NonVerifyingJWTSigner


app = fastapi.FastAPI(
    title="dyff",
)
# api.app.dependency_overrides[api.verify_api_token] = _not_kubernetes
app.mount("/dyff/v0", api.app)
app.mount("/web", web_routes.app)
# apps.site.app.dependency_overrides[apps.site.verify_api_token] = _not_kubernetes


@app.get(f"/health")
async def health() -> int:
    """Required health check for GKE Ingress. Should be specified in the
    k8s Deployment as a readinessProbe.
    """
    return fastapi.status.HTTP_200_OK



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import copy
import json
import math
import time
import urllib
from pathlib import Path
from typing import Dict, List

import requests
from absl import logging

from dyff.api.data import content_hash
from dyff.api.data.sinks import DataSink


class ZenodoRecord:
    def __init__(self, url: str):
        if not url.startswith("https://zenodo.org/api/records/"):
            raise ValueError(
                f"url: expected Zenodo API 'records' endpoint (https://zenodo.org/api/records/)"
            )
        parsed_url = urllib.parse.urlparse(url)  # Verify format
        self._url = url
        r = requests.get(url)
        r.raise_for_status()
        self._record = json.loads(r.content)

    @property
    def resource_id(self) -> str:
        """The unique ID of this resource.

        A resource may consist of multiple files or subdirectories, but they
        can all be found under the ``resource_id``.
        """
        return content_hash(self._record["doi"].encode())

    def metadata_as_json(self) -> Dict:
        """The implementation-specific resource metadata in JSON format.

        Returns:
          A Dict suitable for encoding as JSON.
        """
        return copy.deepcopy(self._record)

    def fetch_into(self, sink: DataSink, *, retries: int = 5) -> List[Path]:
        """Fetches all the data in this source to the given sink.

        Args:
          sink: The DataSink instance
          retries: Must be >= 0. If > 0, the fetch will be retried up to this
            many times, with exponentially-increasing waits between each
            retry up to a maximum wait. The retry limit applies to each upload
            operation separately, but an exception is raised the first time
            any operation exceeds the retry limit.

        Raises:
          Exception: If the upload raises an exception and the retry limit is
            exceeded, the upload exception is propagated to the caller.
        """
        if retries < 0:
            raise ValueError("retries must be non-negative")

        retries_used = 0
        pending_resources = copy.deepcopy(self._record["files"])
        logging.info(
            f"pending_resources: {[fd['links']['self'] for fd in pending_resources]}"
        )
        files = []

        while pending_resources:
            fd = pending_resources[-1]
            url = fd["links"]["self"]
            parsed_url = urllib.parse.urlparse(url)
            filename = parsed_url.path.split("/")[-1]
            sink_path = f"{self.resource_id}/{filename}"
            # The Zenodo checksum looks like 'md5:<hexdigest>'
            algorithm, md5_hex = fd["checksum"].split(":")
            assert algorithm == "md5"

            if sink.resource_exists(sink_path):
                logging.info(f"skipping: {url}")
                pending_resources.pop()
                files.append(sink_path)
            else:
                try:
                    logging.info(f"fetching: {url}")
                    sink.upload_from_http(url, sink_path, md5_hex=md5_hex)
                    pending_resources.pop()
                    files.append(sink_path)
                    retries_used = 0
                except Exception:
                    if retries_used >= retries:
                        raise
                    else:
                        # Limited exponential backoff
                        backoff_exponent = min(retries_used, 5)
                        time.sleep(math.pow(2.0, backoff_exponent))
                        retries_used += 1

        # TODO: Make Builder for metadata object
        meta = {
            "source": self._url,
            "source_kind": "ZenodoRecord",
            "source_metadata": self.metadata_as_json(),
            "resource_id": self.resource_id,
            "sink_id": sink.sink_id,
            "files": [str(path) for path in files],
        }
        sink.upload_string(
            json.dumps(meta), f"{self.resource_id}/_alignmentlabs-meta.json"
        )

        return files



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import absl.app
import absl.flags

from ..data import ingest

# -----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "dataset_yaml", None, "YAML file describing how to parse the dataset."
)
absl.flags.mark_flag_as_required("dataset_yaml")

absl.flags.DEFINE_bool(
    "dryrun",
    False,
    "Parse the YAML file and check paths, etc., but don't run anything.",
)

# -----------------------------------------------------------------------------


def main(unused_argv):
    ingest.ingest(FLAGS.dataset_yaml, dryrun=FLAGS.dryrun)


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import subprocess
import tarfile
from pathlib import Path

import absl.app
import absl.flags
import requests
import ruamel.yaml
from absl import logging
from ruamel.yaml.compat import StringIO as YAMLStringIO

from dyff.schema.platform import Audit

from ...client import Client
from ..sanitize import sanitize_relative_file_path

# ----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "audit_yaml", None, "Path to a YAML file containing the Audit manifest."
)
absl.flags.DEFINE_string(
    "jupyterbook_path", None, "Path to root directory of jupyter book source code."
)
absl.flags.mark_flags_as_required(["audit_yaml", "jupyterbook_path"])

absl.flags.DEFINE_integer(
    "proxy_port",
    8080,
    "Send a request to http://localhost:{proxy_port}/terminate as the last step of execution.",
)

# ----------------------------------------------------------------------------


def main(_unused_argv):
    logging.set_verbosity(logging.INFO)

    dyff = Client()

    yaml = ruamel.yaml.YAML()
    with open(FLAGS.audit_yaml, "r") as fin:
        audit_yaml = yaml.load(fin)
    yaml_string = YAMLStringIO()
    yaml.dump(audit_yaml, yaml_string)
    logging.info(f"audit_yaml:\n{yaml_string.getvalue()}")

    audit = Audit.parse_obj(audit_yaml["spec"])
    jupyterbook_path = Path(FLAGS.jupyterbook_path)

    # Extract the .tar.gz archive
    data = dyff.auditprocedures.download(audit.auditProcedure)
    with tarfile.open(fileobj=data, mode="r:gz") as tgz:
        for member in tgz:
            path = sanitize_relative_file_path(member.name)
            member_file = tgz.extractfile(str(path))
            with open(jupyterbook_path / path, "wb") as fout:
                fout.write(member_file.read())

    # Note: These next two steps shouldn't be necessary if the jupyterbook
    # source was validated on ingestion, but these are untrusted inputs so we'll
    # program defensively.

    # Clear notebook output
    for nb in jupyterbook_path.rglob("*.ipynb"):
        logging.info(f"clearing output in notebook {nb}")
        subprocess.run(
            ["jupyter", "nbconvert", "--clear-output", "--inplace", str(nb)], check=True
        )

    # Overwrite _config.yml with mandatory settings
    with open(jupyterbook_path / "_config.yml", "r") as fin:
        jb_config = yaml.load(fin)
    execute = jb_config.get("execute", {})
    execute["execute_notebooks"] = "force"
    execute["timeout"] = -1  # No limit (timeout should be imposed by k8s Job)
    jb_config["execute"] = execute
    with open(jupyterbook_path / "_config.yml", "w") as fout:
        yaml.dump(jb_config, fout)

    # Run the audit
    subprocess.run(["jupyter-book", "build", FLAGS.jupyterbook_path], check=True)

    # Construct a .tar.gz archive of the HTML output and upload it
    output_root = Path(FLAGS.jupyterbook_path) / "_build" / "html"
    files = (f for f in output_root.rglob("*") if f.is_file())
    # TODO: replace() currently needed because old-style IDs look like 'account/entity'
    # Can be removed after schema migration.
    archive_path = output_root / f"audit-{audit.id.replace('/', '-')}.tar.gz"
    with tarfile.open(archive_path, "w:gz") as tgz:
        for f in files:
            logging.info(f"copying {f}")
            relative_path = f.relative_to(output_root)
            with open(f, "rb") as fin:
                tarinfo = tgz.gettarinfo(arcname=relative_path, fileobj=fin)
                tgz.addfile(tarinfo, fin)
    dyff.audits.upload(key=audit.id, file_path=str(archive_path))

    # Explicitly terminate the sidecar proxy
    try:
        requests.post(f"http://localhost:{FLAGS.proxy_port}/terminate")
    except:
        pass


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import datetime
import time

from dyff.client import Client
from dyff.schema.requests import InferenceSessionCreateRequest

API_KEY: str = ...
ACCOUNT: str = ...

dyffapi = Client(api_key=API_KEY)

session_request = InferenceSessionCreateRequest(
    account=ACCOUNT,
    # databricks/dolly-v2-3b
    inferenceService="ba4ba5c26c9246ee88e127d37cdf548d",
    expires=datetime.datetime.utcnow() + datetime.timedelta(days=1),
    replicas=1,
    useSpotPods=True,
)

session_and_token = dyffapi.inferencesessions.create(session_request)
session = session_and_token.inferencesession
session_id = session.id
session_token = session_and_token.token
print(f"created session:\n{session_and_token.inferencesession}")

# Starting the session can take some time, especially if you requested a GPU
# You can poll the inferencesessions.ready() endpoint to find out if the
# session is ready to accept requests. It will return status 200 if the
# session is ready, and will raise an HttpResponseError with status 503
# (ServiceUnavailable) if the session is not ready. (It may also return 404
# if the session was created recently.)
while not dyffapi.inferencesessions.ready(session_id):
    print(f"[{datetime.datetime.utcnow()}]: not ready")
    # Always use a short sleep when polling in a loop. ready() will usually
    # block for some time as well, but it depends on the runner implementation
    time.sleep(1)
print("Ready")

# If you already have a running session:
# session = dyffapi.inferencesessions.get(session_id)

# Create an inference client using the default interface specified for the
# InferenceService that's being run in the session
interface = session.inferenceService.interface
inference_client = dyffapi.inferencesessions.client(
    session_id,
    session_token,
    # If you don't specify 'interface', the client will use the native JSON
    # interface of the model.
    interface=interface,
    # You can also set any of these separately; they will override the
    # corresponding setting in 'interface' if you specify both.
    # endpoint=interface.endpoint,
    # input_adapter=create_pipeline(interface.inputPipeline),
    # output_adapter=create_pipeline(interface.outputPipeline),
)

# The input is {"text": ...} because the default interface for the dolly-v2-3b
# service maps {"text": ...} -> {"prompt": ...}
y = inference_client.infer({"text": "Open the pod bay doors, Hal!"})
print(y)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import math
from collections import Counter
from typing import NamedTuple, Optional

import absl.app
import absl.flags
import pyarrow
import ruamel.yaml
from absl import logging
from ruamel.yaml.compat import StringIO as YAMLStringIO

from dyff.api import storage
from dyff.schema import ids
from dyff.schema.dataset import arrow

# -----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "evaluation_yaml", None, "Path to a YAML file containing the Evaluation manifest."
)


# Return codes should be combined with bitwise-or
RETURNCODE_SUCCESS = 0
RETURNCODE_OUTPUT_ERROR = 1 << 0
RETURNCODE_MISSING = 1 << 1
RETURNCODE_DUPLICATES = 1 << 2

# -----------------------------------------------------------------------------


class ItemID(NamedTuple):
    replication: str
    item_index: int


def verify_output(evaluation, *, repartition: bool) -> int:
    evaluation_id = evaluation["spec"]["id"]
    dataset = evaluation["spec"]["dataset"]
    replications: int = evaluation["spec"]["replications"]

    outputs_dataset = arrow.open_dataset(storage.paths.outputs_raw(evaluation_id))
    feature_schema = outputs_dataset.schema
    num_rows = outputs_dataset.count_rows()
    output_size = storage.storage_size(storage.paths.outputs_raw(evaluation_id))
    size_MBi = output_size / (1000 * 1024)
    MBi_per_row = size_MBi / num_rows
    # Target is 100MBi per file
    n = 100.0 / MBi_per_row
    # Nice round number -- smallest power of 2 s.t. n * MBi_per_row >= 100
    rows_per_file = int(math.pow(2, math.ceil(math.log2(n))))
    logging.info(
        f"rows: {num_rows}; size (MBi): {size_MBi}; rows_per_file: {rows_per_file}"
    )

    input_dataset = arrow.open_dataset(storage.paths.dataset_root(dataset))
    filter_expression = None
    # if (dataset_config := evaluation["spec"].get("datasetConfiguration")) is not None:
    #     if (filters := dataset_config.get("filters")) is not None:
    #         filter_expression = parse_dataset_filters(filters)
    #         logging.info(f"dataset filter: {filter_expression}")
    input_indices = []
    for b in input_dataset.to_batches(columns=["_index_"], filter=filter_expression):
        input_indices.extend(b.to_pandas()["_index_"])
    unique_input_indices = set(input_indices)
    assert len(unique_input_indices) == len(
        input_indices
    ), "duplicate indices in input dataset"
    logging.info(f"input dataset: {len(unique_input_indices)} unique indices")
    replication_ids = [
        ids.replication_id(evaluation_id, i) for i in range(replications)
    ]
    input_items = []
    for index in input_indices:
        for replication in replication_ids:
            input_items.append(ItemID(replication=replication, item_index=index))

    partition_keys = evaluation["spec"].get("outputDatasetPartitions")
    if partition_keys:
        # If these fields are present, we want to represent them as partitions
        partition_fields = []
        for partition_key in partition_keys:
            field_index = feature_schema.get_field_index(partition_key)
            if field_index != -1:
                partition_fields.append(feature_schema.field(field_index))
        partition_schema = (
            pyarrow.schema(partition_fields) if partition_fields else None
        )
    else:
        # FIXME: That PyArrow bug seems to be back where it doesn't correctly
        # populate .partitioning.schema when there isn't actually a partitioning
        partition_schema = None
        # Else keep the partitioning from the unverified outputs
        # partition_schema = (
        #     outputs_dataset.partitioning and outputs_dataset.partitioning.schema
        # )
    logging.info(f"feature_schema:\n{feature_schema}")
    logging.info(f"partition_schema:\n{partition_schema}")

    processed: list[ItemID] = []
    errors: list[ItemID] = []

    def output_generator(outputs_dataset, processed: list[ItemID] = processed):
        for b in outputs_dataset.to_batches():
            batch = []
            for item in b.to_pylist():
                # TODO: We're not doing any content checking anymore because
                # all we have is the arrow schema
                item_id = ItemID(
                    replication=item["_replication_"], item_index=item["_index_"]
                )
                batch.append(item)
                processed.append(item_id)
                if len(processed) % 100000 == 0:
                    logging.info(f"processed {len(processed)} items")
            yield pyarrow.RecordBatch.from_pylist(batch, schema=feature_schema)

    outputs_dataset = arrow.open_dataset(storage.paths.outputs_raw(evaluation_id))

    arrow.write_dataset(
        output_generator(outputs_dataset),
        output_path=storage.paths.outputs_verified(evaluation_id),
        feature_schema=feature_schema,
        partition_schema=partition_schema,
        max_rows_per_file=rows_per_file,
        # max_rows_per_group must be less than or equal to max_rows_per_file
        max_rows_per_group=rows_per_file,
    )

    return_code = 0
    logging.info(f"final: {len(processed)} items; {len(set(processed))} unique")
    if len(errors) > 0:
        logging.error(f"found output errors: {errors}")
        return_code |= RETURNCODE_OUTPUT_ERROR
    counts = Counter(processed)
    duplicates = [k for k, v in counts.items() if v > 1]
    if len(duplicates) > 0:
        logging.error(f"found duplicate items: {duplicates}")
        return_code |= RETURNCODE_DUPLICATES
    missing = set(input_items).difference(set(processed))
    if len(missing) > 0:
        logging.error(f"missing items: {missing}")
        return_code |= RETURNCODE_MISSING
    return return_code


def main(_unused_argv: list[str]) -> Optional[int]:
    logging.set_verbosity(logging.INFO)

    yaml = ruamel.yaml.YAML()
    with open(FLAGS.evaluation_yaml, "r") as fin:
        evaluation = yaml.load(fin)
    yaml_string = YAMLStringIO()
    yaml.dump(evaluation, yaml_string)
    logging.info(f"evaluation_yaml:\n{yaml_string.getvalue()}")

    return verify_output(evaluation, repartition=True)


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os

from dyff.client.client import Client
from dyff.schema.platform import (
    ModelArtifact,
    ModelArtifactHuggingFaceCache,
    ModelArtifactKind,
    ModelResources,
    ModelSource,
    ModelSourceHuggingFaceHub,
    ModelSourceKinds,
    ModelStorage,
    ModelStorageMedium,
)
from dyff.schema.requests import ModelCreateRequest

API_KEY: str = os.environ["DYFF_API_KEY"]
ACCOUNT: str = "public"

dyffapi = Client(api_key=API_KEY)

model_request = ModelCreateRequest(
    account=ACCOUNT,
    name="databricks/dolly-v2-3b",
    artifact=ModelArtifact(
        kind=ModelArtifactKind.HuggingFaceCache,
        huggingFaceCache=ModelArtifactHuggingFaceCache(
            repoID="databricks/dolly-v2-3b",
            revision="f6c9be08f16fe4d3a719bee0a4a7c7415b5c65df",
        ),
    ),
    source=ModelSource(
        kind=ModelSourceKinds.HuggingFaceHub,
        huggingFaceHub=ModelSourceHuggingFaceHub(
            repoID="databricks/dolly-v2-3b",
            revision="f6c9be08f16fe4d3a719bee0a4a7c7415b5c65df",
        ),
    ),
    storage=ModelStorage(
        medium=ModelStorageMedium.PersistentVolume,
    ),
    resources=ModelResources(
        storage="10Gi",
        memory="16Gi",
    ),
)

model = dyffapi.models.create(model_request)
print(model)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""Web server for the dyff platform.
"""

import logging
from contextlib import asynccontextmanager

import fastapi

from ..core.config import config
from . import api


class EndpointFilter(logging.Filter):
    def __init__(self, path: str, *args, verb: str = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._path = path
        self._verb = verb

    def filter(self, record):
        (
            remote_address,
            verb,
            query_string,
            html_version,
            status_code,
            *rest,
        ) = record.args
        return not all(
            [
                self._verb is None or verb == self._verb,
                query_string.startswith(self._path),
            ]
        )


try:
    import uvicorn.config

    # Filter out log messages about the '/health' endpoint, because they happen
    # a lot and they're not interesting.
    uvicorn_logger = logging.getLogger("uvicorn.access")
    uvicorn_logger.addFilter(EndpointFilter(path="/health"))
    uvicorn.config.LOGGING_CONFIG["loggers"]["api-server"] = {
        "handlers": ["default"],
        "level": logging.INFO,
    }
except Exception:
    pass


@asynccontextmanager
async def _lifespan(app: fastapi.FastAPI):
    # FIXME: Try as I might, I can't get logging to print anything unless I go
    # through the 'uvicorn' logger. Since all I really want to do is see the
    # configuration, I'm just going to use a print() for now.
    # log = logging.getLogger("api-server")
    # log.info(f"aStarting API server with configuration:\n{config.json(indent=2)}")
    # log = logging.getLogger("uvicorn")
    # log.info(f"bStarting API server with configuration:\n{config.json(indent=2)}")
    print(
        f"Starting API server with configuration:\n{config.json(indent=2)}", flush=True
    )
    yield


app = fastapi.FastAPI(
    title="dyff",
    lifespan=_lifespan,
)
app.mount("/dyff/v0", api.app)


@app.get(f"/health")
async def health() -> int:
    """Required health check for GKE Ingress. Should be specified in the
    k8s Deployment as a readinessProbe.
    """
    return fastapi.status.HTTP_200_OK



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import json
import os

from dyff.client.client import Client
from dyff.schema.platform import (
    Accelerator,
    AcceleratorGPU,
    DataSchema,
    DyffDataSchema,
    InferenceInterface,
    InferenceServiceRunner,
    InferenceServiceRunnerKind,
    ModelResources,
    SchemaAdapter,
)
from dyff.schema.requests import InferenceServiceCreateRequest

API_KEY: str = os.environ["DYFF_API_KEY"]
ACCOUNT: str = "public"

dyffapi = Client(api_key=API_KEY)

model = dyffapi.models.query(name="tiiuae/falcon-7b", account="public")
assert len(model) == 1 and model[0].id == "371288ec69724bf8bebf51811c581f6b"
model_id = model[0].id

service_request = InferenceServiceCreateRequest(
    account=ACCOUNT,
    name="tiiuae/falcon-7b/default",
    model=model_id,
    runner=InferenceServiceRunner(
        kind=InferenceServiceRunnerKind.VLLM,
        # T4 GPUs don't support bfloat format, so force standard float format
        args=[
            "--dtype",
            "float16",
        ],
        accelerator=Accelerator(
            kind="GPU",
            gpu=AcceleratorGPU(
                hardwareTypes=["nvidia.com/gpu-a100"],
                memory="16Gi",
            ),
        ),
        resources=ModelResources(
            storage="20Gi",
            memory="16Gi",
        ),
    ),
    interface=InferenceInterface(
        # This is the inference endpoint for the vLLM runner
        endpoint="generate",
        # The output records should look like: {"text": "To be, or not to be"}
        outputSchema=DataSchema.make_output_schema(
            DyffDataSchema(
                components=["text.Text"],
            ),
        ),
        # How to convert the input dataset into the format the runner expects
        inputPipeline=[
            # {"text": "The question"} -> {"prompt": "The question"}
            SchemaAdapter(
                kind="TransformJSON",
                configuration={"prompt": "$.text"},
            ),
        ],
        # How to convert the runner output to match outputSchema
        outputPipeline=[
            # {"text": ["The answer"]} -> [{"text": "The answer"}]
            SchemaAdapter(
                kind="ExplodeCollections",
                configuration={"collections": ["text"]},
            ),
        ],
    ),
)

print(json.dumps(service_request.dict(), indent=2))

service = dyffapi.inferenceservices.create(service_request)
print(f"created inferenceservice:\n{service}")



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import functools
from typing import Optional

from dyff.schema.platform import Artifact, StorageSignedURL

from ...api.backend.base.storage import StorageBackend
from . import paths


@functools.lru_cache()
def _get_backend() -> StorageBackend:
    # FIXME: Instantiate backend from config
    from dyff.api.backend.gcloud.storage import GCloudStorageBackend

    return GCloudStorageBackend()


def storage_size(path: str) -> int:
    return _get_backend().storage_size(path)


def list_dir(path: str) -> list[str]:
    return _get_backend().list_dir(path)


def download_recursive(source: str, destination: str) -> None:
    return _get_backend().download_recursive(source, destination)


def upload_recursive(source: str, destination: str) -> None:
    return _get_backend().upload_recursive(source, destination)


def signed_url_for_dataset_upload(
    dataset_id: str,
    artifact: Artifact,
    *,
    size_limit_bytes: Optional[int] = None,
    storage_path: Optional[str] = None,
) -> StorageSignedURL:
    return _get_backend().signed_url_for_dataset_upload(
        dataset_id,
        artifact,
        size_limit_bytes=size_limit_bytes,
        storage_path=storage_path,
    )


def dataset_artifact_md5hash(
    dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
) -> bytes:
    return _get_backend().dataset_artifact_md5hash(
        dataset_id, artifact_path, storage_path=storage_path
    )


__all__ = [
    "paths",
    "dataset_artifact_md5hash",
    "download_recursive",
    "list_dir",
    "signed_url_for_dataset_upload",
    "storage_size",
]



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os

import kubernetes.config


def load_config():
    if not hasattr(load_config, "_loaded"):
        # https://stackoverflow.com/a/51035715
        if os.environ.get("KUBERNETES_SERVICE_HOST"):
            kubernetes.config.load_incluster_config()
        else:
            kubernetes.config.load_kube_config()
        load_config._loaded = True


load_config()


__all__ = ["load_config"]



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import click
import yaml

from dyff.api.backend.base.auth import AuthBackend
from dyff.api.tokens import (
    ACCOUNT_TOKEN,
    TOKEN_CHOICES,
    Signer,
    generate_api_key,
    get_signer,
    hashed_api_key,
)
from dyff.core import dynamic_import
from dyff.core.config import config
from dyff.schema.platform import AccessGrant, Entities

from . import roles


@click.group()
def tokens():
    pass


def get_comma_separated_list(ctx, _param, value) -> list[str]:
    if value is None:
        return []
    values = value.split(",")
    return values


@tokens.command()
@click.option(
    "token_type",
    "--type",
    "-t",
    type=click.Choice(TOKEN_CHOICES),
    required=True,
    help="Type of API key to create.",
)
@click.option(
    "account_id",
    "--account-id",
    "-i",
    metavar="ID",
    type=str,
    help="Account ID to associate token with.",
)
@click.option(
    "resources",
    "--resources",
    "-r",
    type=str,
    callback=get_comma_separated_list,
    help="Resources to grant access to.",
)
@click.option(
    "functions",
    "--functions",
    "-f",
    type=str,
    callback=get_comma_separated_list,
    help="Functions to grant access to.",
)
@click.option(
    "accounts",
    "--accounts",
    "-a",
    type=str,
    callback=get_comma_separated_list,
    help="Accounts to grant access to.",
)
@click.option(
    "entities",
    "--entities",
    "-e",
    type=str,
    callback=get_comma_separated_list,
    help="Entities to grant access to.",
)
@click.option(
    "config_file",
    "--config",
    "-c",
    type=click.File(),
    help="Access grant config file specified with JSON / YAML.",
)
@click.option("role", "--role", type=str, help="Named Role specifying access grants")
def create(
    role: str,
    config_file: str,
    account_id: str,
    token_type: str,
    resources: list[str],
    functions: list[str],
    accounts: list[str],
    entities: list[str],
):
    auth_backend: AuthBackend = dynamic_import.instantiate(config.api.auth.backend)

    if token_type != ACCOUNT_TOKEN:
        raise NotImplementedError("ephemeral token support has not yet been added.")

    if role and config_file:
        raise click.ClickException("--role and --config are mutually exclusive")

    if role:
        spec = roles.named_role(role, account_id)
        account_id = spec["account"]
        grants = spec["grants"]
    elif config_file:
        spec = yaml.safe_load(config_file)
        account_id = spec["account"]
        grants = [AccessGrant.parse_obj(grant) for grant in spec["grants"]]
    else:
        if account_id is None:
            raise click.UsageError("--account-id/-a must be specified")
        grants = [
            AccessGrant.parse_obj(
                dict(
                    accounts=accounts,
                    entities=entities,
                    functions=functions,
                    resources=resources,
                )
            )
        ]

    auth_backend.get_account(id=account_id)

    signer: Signer = get_signer(
        config.api.auth.api_key_signing_secret.get_secret_value()
    )
    api_key = generate_api_key(
        subject_type=Entities.Account,
        subject_id=account_id,
        grants=grants,
        generate_secret=True,
    )
    hashed_key = hashed_api_key(api_key)
    signed_key = signer.sign_api_key(api_key)

    auth_backend.revoke_all_api_keys(account_id)
    auth_backend.add_api_key(account_id, hashed_key)
    click.echo(signed_key)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import List, Optional

import pymongo
import pymongo.read_concern
import pymongo.read_preferences
import pymongo.write_concern
from pymongo.client_session import ClientSession

from dyff.schema import ids
from dyff.schema.platform import APIKey

from ....core.config import config
from ... import timestamp
from ...exceptions import EntityExistsError
from ..base.auth import Account, AuthBackend


class MongoDBAuthBackend(AuthBackend):
    def __init__(self):
        connection_string = config.api.auth.mongodb.connection_string
        self._client = pymongo.MongoClient(connection_string.get_secret_value())

        # Interact with the database in a way that gives strong consistency
        # (These are currently the default settings; I prefer to be explicit)
        self._accounts_db = self._client.get_database(
            config.api.auth.mongodb.database,
            read_concern=pymongo.read_concern.ReadConcern("majority"),
            read_preference=pymongo.ReadPreference.PRIMARY,
            write_concern=pymongo.write_concern.WriteConcern("majority", wtimeout=5000),
        )

    def _get_account(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
        session: ClientSession = None,
    ) -> Optional[Account]:
        filter = {}
        if id is not None:
            filter["_id"] = id
        if name is not None:
            filter["name"] = name
        if len(filter) == 0:
            raise ValueError("must specify at least one of {id, name}")
        result = self._accounts_db.accounts.find_one(filter, session=session)
        if result:
            result = dict(result)
            result["id"] = result["_id"]
            del result["_id"]
            return Account.parse_obj(result)
        return None

    def _insert_account(
        self, account: Account, *, session: ClientSession = None
    ) -> None:
        d = account.dict()
        d["_id"] = d["id"]
        del d["id"]
        self._accounts_db.accounts.insert_one(d, session=session)

    def create_account(self, name: str) -> Account:
        """Create a new Account.

        Parameters:
        name: A unique human-readable name

        Returns:
        Entity representing the Account.
        """
        duplicate_name = self._get_account(name=name)
        if duplicate_name is not None:
            raise EntityExistsError("duplicate account name")
        id = ids.generate_entity_id()
        account = Account(id=id, name=name, creationTime=timestamp.now())
        self._insert_account(account)
        return account

    def delete_account(self, account_id: str):
        """Delete an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        raise NotImplementedError()

    def get_account(
        self, *, id: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Account]:
        """Get Account by ID.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        return self._get_account(id=id, name=name)

    def add_api_key(self, account_id: str, api_key: APIKey) -> None:
        """Add a new APIKey to an Account.

        Raises ``ValueError`` if an APIKey with the same ID already exists.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The new API key.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        if any(account_key.id == api_key.id for account_key in account.apiKeys):
            raise ValueError(f"APIKey with .id {api_key.id} already exists")
        # Don't have to translate APIKey.id because APIKey isn't a
        # top-level entity
        self._accounts_db.accounts.update_one(
            {"_id": account_id}, {"$push": {"apiKeys": api_key.dict()}}
        )

    def revoke_api_key(self, account_id: str, api_key_id: str) -> None:
        """Revoke an APIKey associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The unique identifier of the APIKey.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        filtered = [api_key for api_key in account.apiKeys if api_key.id != api_key_id]
        if len(filtered) == len(account.apiKeys):
            raise ValueError(f"no APIKey with .id {api_key_id} in account")
        # Don't have to translate APIKey.id because APIKey isn't a
        # top-level entity
        self._accounts_db.accounts.update_one(
            {"_id": account_id},
            {"$set": {"apiKeys": [key.dict() for key in filtered]}},
        )

    def revoke_all_api_keys(self, account_id: str) -> None:
        """Revoke all API keys for the given Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        result = self._accounts_db["accounts"].update_one(
            {"_id": account_id}, {"$set": {"apiKeys": []}}
        )
        if result.matched_count != 1:
            raise ValueError(f"no Account with .id {account_id}")

    def get_api_key(self, account_id: str, api_key_id: str) -> Optional[APIKey]:
        """Get an APIKey associated with an Account by ID.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key_id: The unique identifier of the APIKey.
        """
        api_keys = self.get_all_api_keys(account_id)
        for api_key in api_keys:
            if api_key.id == api_key_id:
                return api_key
        return None

    def get_all_api_keys(self, account_id: str) -> List[APIKey]:
        """Get the API keys associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """
        account = self._get_account(id=account_id)
        if account is None:
            raise ValueError(f"no Account with .id {account_id}")
        account = Account.parse_obj(account)
        return account.apiKeys



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import absl.app
import absl.flags
import google.cloud.devtools.cloudbuild_v1 as gcb
import google.cloud.devtools.cloudbuild_v1.types as gcb_types

FLAGS = absl.flags.FLAGS


absl.flags.DEFINE_string("project", None, "The Google Cloud project ID.")
absl.flags.mark_flag_as_required("project")

absl.flags.DEFINE_string(
    "location", None, "The Google Cloud location of the Artifact Registry."
)
absl.flags.mark_flag_as_required("location")

absl.flags.DEFINE_string(
    "source_bucket", None, "The GCS bucket containing the model source."
)
absl.flags.mark_flag_as_required("source_bucket")

absl.flags.DEFINE_string("account", None, "The account identifier of the model owner.")
absl.flags.mark_flag_as_required("account")

absl.flags.DEFINE_string("model", None, "The ID of the model.")
absl.flags.mark_flag_as_required("model")

absl.flags.DEFINE_string("service", None, "The name of the ModelService to build.")
absl.flags.mark_flag_as_required("service")


def main(unused_argv):
    image = f"{FLAGS.location}-docker.pkg.dev/{FLAGS.project}/dyff-models/{FLAGS.account}/{FLAGS.model}"
    build = gcb.Build(
        steps=[
            gcb_types.BuildStep(
                name="gcr.io/cloud-builders/docker",
                args=[
                    "buildx",
                    "build",
                    f"--tag={image}",
                    "--file=env/docker/Dockerfile",
                    ".",
                ],
            )
        ],
        source=gcb_types.Source(
            storage_source=gcb_types.StorageSource(
                bucket=FLAGS.source_bucket,
                object_=f"{FLAGS.account}/{FLAGS.model}/{FLAGS.service}.tar.gz",
            )
        ),
        images=[image],
    )

    client = gcb.CloudBuildClient()
    operation = client.create_build(project_id=FLAGS.project, build=build)
    response = operation.result()
    return int(response.status != gcb_types.Build.Status.SUCCESS)


if __name__ == "__main__":
    absl.app.run(main)
